"""Osc.sc"""

from ...base import builtins as bi
from .. import ugen as ugn
from .. import _graphparam as gpp
from . import noise as nse
from . import pan


class Osc(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, bufnum, freq=440.0, phase=0.0):
        return cls._multi_new('audio', bufnum, freq, phase)

    @classmethod
    def kr(cls, bufnum, freq=440.0, phase=0.0):
        return cls._multi_new('control', bufnum, freq, phase)


class SinOsc(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, freq=440.0, phase=0.0):
        return cls._multi_new('audio', freq, phase)

    @classmethod
    def kr(cls, freq=440.0, phase=0.0):
        return cls._multi_new('control', freq, phase)


class SinOscFB(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, freq=440.0, feedback=0.0):
        return cls._multi_new('audio', freq, feedback)

    @classmethod
    def kr(cls, freq=440.0, feedback=0.0):
        return cls._multi_new('control', freq, feedback)


class OscN(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, bufnum, freq=440.0, phase=0.0):
        return cls._multi_new('audio', bufnum, freq, phase)

    @classmethod
    def kr(cls, bufnum, freq=440.0, phase=0.0):
        return cls._multi_new('control', bufnum, freq, phase)


class VOsc(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, bufpos, freq=440.0, phase=0.0):
        return cls._multi_new('audio', bufpos, freq, phase)

    @classmethod
    def kr(cls, bufpos, freq=440.0, phase=0.0):
        return cls._multi_new('control', bufpos, freq, phase)


class VOsc3(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, bufpos, freq1=110.0, freq2=220.0, freq3=440.0):
        return cls._multi_new('audio', bufpos, freq1, freq2, freq3)

    @classmethod
    def kr(cls, bufpos, freq1=110.0, freq2=220.0, freq3=440.0):
        return cls._multi_new('control', bufpos, freq1, freq2, freq3)


class COsc(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, bufnum, freq=440.0, beats=0.5):
        return cls._multi_new('audio', bufnum, freq, beats)

    @classmethod
    def kr(cls, bufnum, freq=440.0, beats=0.5):
        return cls._multi_new('control', bufnum, freq, beats)


class Formant(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, fundfreq=440.0, formfreq=1760.0, bwfreq=880.0):
        return cls._multi_new('audio', fundfreq, formfreq, bwfreq)


class LFSaw(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, freq=440.0, iphase=0.0):
        return cls._multi_new('audio', freq, iphase)

    @classmethod
    def kr(cls, freq=440.0, iphase=0.0):
        return cls._multi_new('control', freq, iphase)


class LFPar(LFSaw):
    pass


class LFCub(LFSaw):
    pass


class LFTri(LFSaw):
    pass


class LFGauss(ugn.UGen):
    @classmethod
    def ar(cls, duration=1, width=0.1, iphase=0.0, loop=1, done_action=0):
        return cls._multi_new('audio', duration, width, iphase, loop, done_action)

    @classmethod
    def kr(cls, duration=1, width=0.1, iphase=0.0, loop=1, done_action=0):
        return cls._multi_new('control', duration, width, iphase, loop, done_action)

    @property
    def _minval(self):
        width = self.inputs[1]
        return bi.exp(1.0 / (-2.0 * bi.squared(width)))

    def range(self, min=0, max=1):
        return self.linlin(self._minval, 1, min, max)


class LFPulse(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, freq=440.0, iphase=0.0, width=0.5):
        return cls._multi_new('audio', freq, iphase, width)

    @classmethod
    def kr(cls, freq=440.0, iphase=0.0, width=0.5):
        return cls._multi_new('control', freq, iphase, width)

    @classmethod
    def signal_range(cls):  # override
        return 'unipolar'


class VarSaw(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, freq=440.0, iphase=0.0, width=0.5):
        return cls._multi_new('audio', freq, iphase, width)

    @classmethod
    def kr(cls, freq=440.0, iphase=0.0, width=0.5):
        return cls._multi_new('control', freq, iphase, width)


class Impulse(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, freq=440.0, phase=0.0):
        return cls._multi_new('audio', freq, phase)

    @classmethod
    def kr(cls, freq=440.0, phase=0.0):
        return cls._multi_new('control', freq, phase)

    @classmethod
    def signal_range(cls):  # override
        return 'unipolar'


class SyncSaw(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, sync_freq=440.0, saw_freq=440.0):
        return cls._multi_new('audio', sync_freq, saw_freq)

    @classmethod
    def kr(cls, sync_freq=440.0, saw_freq=440.0):
        return cls._multi_new('control', sync_freq, saw_freq)


# TPulse


class Index(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, bufnum, input=0.0):
        return cls._multi_new('audio', bufnum, input)

    @classmethod
    def kr(cls, bufnum, input=0.0):
        return cls._multi_new('control', bufnum, input)


class WrapIndex(Index):
    pass


class IndexInBetween(Index):
    pass


class DetectIndex(Index):
    pass


class Shaper(Index):
    pass


class IndexL(Index):
    pass


class DegreeToKey(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, bufnum, input=0.0, octave=12.0):
        return cls._multi_new('audio', bufnum, input, octave)

    @classmethod
    def kr(cls, bufnum, input=0.0, octave=12.0):
        return cls._multi_new('control', bufnum, input, octave)


class Select(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, which, lst):
        return cls._multi_new('audio', which, *lst)

    @classmethod
    def kr(cls, which, lst):
        return cls._multi_new('control', which, *lst)

    def _check_inputs(self):
        if self.rate == 'audio':
            for i in range(1, len(self.inputs)):
                if gpp.ugen_param(self.inputs[i])._as_ugen_rate() != 'audio':
                    return f'input was not audio rate: {self.inputs[i]}'
        return self._check_valid_inputs()


# Pseudo UGens don't return instances of themselves but of another subclass
# of UGen, ChannelList or (maybe) optimizations (UGenParameter scalar types).

class SelectX(ugn.PseudoUGen):
    @classmethod
    def _new1(cls, rate, which, lst, wrap):  # wrap was never implemented (needs to be a proper UGen)
        selector = cls._method_selector_for_rate(rate)
        return getattr(cls._crossfade_class(), selector)(
            getattr(Select, selector)(bi.round(which, 2), lst),
            getattr(Select, selector)(bi.trunc(which, 2) + 1, lst),
            bi.fold2(which * 2 - 1, 1))

    @classmethod
    def ar(cls, which, lst, wrap=1):
        return cls._new1('audio', which, lst, wrap)

    @classmethod
    def kr(cls, which, lst, wrap=1):
        return cls._new1('control', which, lst, wrap)

    @classmethod
    def _method_selector_for_rate(cls, rate):  # FIXME: API, same in LinLin
        if rate == 'audio':
            return 'ar'
        elif rate == 'control':
            return 'kr'
        raise AttributeError(f'{cls.__name__} has no {rate} rate constructor')

    @classmethod
    def _crossfade_class(cls):
        return pan.XFade2


class LinSelectX(SelectX):
    @classmethod
    def _crossfade_class(cls):
        return pan.LinXFade2


class SelectXFocus(ugn.PseudoUGen):
    @classmethod
    def new(cls, which, lst, focus=1, wrap=False):
        if wrap:
            return ugn.Mix.new(
                [bi.max(1 - (bi.moddif(which, i, len(lst)) * focus), 0) * input
                 for i, input in enumerate(lst)])
        else:
            return ugn.Mix.new(
                [bi.max(1 - (bi.absdif(which, i) * focus), 0) * input
                 for i, input in enumerate(lst)])

    @classmethod
    def ar(cls, which, lst, focus=1, wrap=False):
        return self.new(which, array, focus, wrap)  # ar/kr are fake rate constructors.

    @classmethod
    def kr(cls, which, lst, focus=1, wrap=False):
        return self.new(which, array, focus, wrap)


class Vibrato(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, freq=440.0, rate=6, depth=0.02, delay=0.0, onset=0.0,
           rate_variation=0.04, depth_variation=0.1, iphase=0.0, trig=0.0):
        return cls._multi_new('audio', freq, rate, depth, delay, onset,
                              rate_variation, depth_variation, iphase, trig)

    @classmethod
    def kr(cls, freq=440.0, rate=6, depth=0.02, delay=0.0, onset=0.0,
           rate_variation=0.04, depth_variation=0.1, iphase=0.0, trig=0.0):
        return cls._multi_new('control', freq, rate, depth, delay, onset,
                              rate_variation, depth_variation, iphase, trig)


class TChoose(ugn.PseudoUGen):
    @classmethod
    def ar(cls, trig, lst):
        self._check_empty_list(lst)
        return Select.ar(nse.TIRand.ar(0, len(lst) - 1, trig), lst)

    @classmethod
    def kr(cls, trig, lst):
        self._check_empty_list(lst)
        return Select.kr(nse.TIRand.kr(0, len(lst) - 1, trig), lst)

    def _check_empty_list(self, lst):
        if len(lst) == 0:
            raise ValueError("TChoose: lst can't be empty")


class TWChoose(ugn.PseudoUGen):
    @classmethod
    def ar(cls, trig, lst, weights, normalize=0):
        return Select.ar(nse.TWindex.ar(trig, weights, normalize), lst)

    @classmethod
    def kr(cls, trig, lst, weights, normalize=0):
        return Select.kr(nse.TWindex.kr(trig, weights, normalize), lst)




from abc import ABC, abstractmethod
import logging
import threading
import atexit
import errno
import struct
import socket
import time
import subprocess
import sys
import pprint

from ..synth import server as srv
from . import clock as clk
from . import stream as stm
from . import _taskq as tsq
from . import main as _libsc3
from . import netaddr as nad
from . import _osclib as oli
from . import functions as fn
from . import platform as plf


__all__ = ['OscUdpInterface', 'OscTcpInterface', 'OscNrtInterface']


_logger = logging.getLogger(__name__)


class OscInterface(ABC):
    def __init__(self, port, port_range):
        self._port = port
        self._port_range = port_range
        self._socket = None
        self._recv_functions = set()

    @property
    def port(self):
        return self._port

    @property
    def port_range(self):
        return self._port_range

    @property
    def socket(self):
        return self._socket

    @property
    def recv_functions(self):
        return self._recv_functions

    def add_recv_func(self, func):
        '''
        Register a callable to be evaluated each time an OSC message arrives.
        Signature of func is:
            msg: OSC message as list.
            time: Time of arrival meassured from main.elapsed_time().
            addr: A NetAddr object with sender's address.
            port: Local port as int.
        '''
        self._recv_functions.add(func)

    def remove_recv_func(self, func):
        '''Unregister func callback.'''
        self._recv_functions.discard(func)

    def recv(self, addr, time, *msg):
        '''
        This method is the handler of all incoming OSC messages or bundles
        to be registered once for each OSC server interface in subclasses.

        Args:
            addr: A tuple (sender_ip:str, sender_port:int).
            time: OSC timetag as 64bits unsigned integer.
            *msg: OSC message as address followed by values.
        '''
        addr = nad.NetAddr(addr[0], addr[1])

        if time is None or time == oli.IMMEDIATELY:
            time = _libsc3.main.elapsed_time()
        else:
            time = clk.SystemClock.osc_to_elapsed_time(time)

        def sched_func():
            for func in self.recv_functions:
                func(list(msg), time, addr, self.port)

        clk.SystemClock.sched(0, sched_func)  # Updates logical time.

    def send_msg(self, target, *args):
        '''
        args are values to create one message.
        target is a tuple (hostname, port).
        sclang converts True to 1, False to 0, None and empty lists to 0.
        Non empty lists are converted to blobs containing osc messages or
        bundles. Empty strings are sent unchanged.
        '''
        self._send(self._build_msg(list(args)), target)

    def send_bundle(self, target, time, *elements):
        '''
        args are lists of values to creae a message or bundle each.
        target is a tuple (hostname, port).
        If time is None the OSC library must send 1 (immediate) as timetag.
        If time is negative it will be substracted from elapsed time and be
        an already late timetag (no check for sign).
        '''
        # Time is calculated in _build_bundle to support
        # sub-bundles relative time by calling _get_timetag().
        self._send(self._build_bundle([time, *elements]), target)

    @abstractmethod
    def _send(self, msg, target):
        pass

    def _build_msg(self, arg_list):
        # ['/path', arg1, arg2, ..., argN]
        msg_builder = oli.OscMessageBuilder(arg_list[0])
        for arg in arg_list[1:]:
            if arg is None:
                msg_builder.add_arg(0)
            elif isinstance(arg, bool):
                msg_builder.add_arg(int(arg))
            elif isinstance(arg, list):
                if len(arg) == 0:
                    msg_builder.add_arg(0)
                elif isinstance(arg[0], str):
                    msg_builder.add_arg(self._build_msg(arg).dgram)
                elif isinstance(arg[0], (int, float, type(None)))\
                and len(arg) > 1 and isinstance(arg[1], list):
                    msg_builder.add_arg(self._build_bundle(arg).dgram)
                else:
                    raise ValueError(
                        'lists within messages must be valid '
                        f'OSC messages or bundles: {arg}')
            elif arg == '[':
                msg_builder.args.append(
                    (msg_builder.ARG_TYPE_ARRAY_START, None))
            elif arg == ']':
                msg_builder.args.append(
                    (msg_builder.ARG_TYPE_ARRAY_STOP, None))
            else:
                msg_builder.add_arg(arg)  # Infiere correctamente el resto de los tipos.
        return msg_builder.build()

    def _build_bundle(self, arg_list):
        # [time, ['/path', arg1, arg2, ..., argN], [...], ...]
        timetag = self._get_timetag(arg_list[0])
        bndl_builder = oli.OscBundleBuilder(timetag)
        for arg in arg_list[1:]:
            if isinstance(arg[0], str):
                bndl_builder.add_content(self._build_msg(arg))
            elif isinstance(arg[0], (int, float, type(None))):
                self._check_subtime(arg_list[0], arg[0])
                bndl_builder.add_content(self._build_bundle(arg))
            else:
                raise ValueError(
                    'elements within bundles must be valid '
                    f'OSC messages or bundles: {arg}')
        return bndl_builder.build()

    @staticmethod
    def _get_timetag(time):
        if time is None or time < 0.0:
            return oli.IMMEDIATELY
        else:
            time += _libsc3.main.current_tt.seconds
            return clk.SystemClock.elapsed_time_to_osc(time)

    @staticmethod
    def _check_subtime(time, subtime):
        # OSC spec. 1.0. This check should be done by _osclib.
        if time is None:
            return
        if subtime is None or time > subtime:
            raise ValueError(
                'nested bundle time must be >= enclosing bundle time')

    def __str__(self):
        return f'{type(self).__name__} port {self._port}'


    ### *** SI SE BORRAN ESTOS MÉTODOS POR SER ESPECÍFICOS DE CASA SUBCLASE
    ### *** HAY QUE TENER REVISAR EL USO DE POLIMORFISMO. LOS NOMBRES DE NRT
    ### *** NO ESTÁN DECIDIDOS AÚN.

    ### UDP Interface ###

    def start(self):
        pass

    def stop(self):
        pass

    def running(self):
        pass


    ### TCP Interface ###

    def bind(self):
        pass

    def connect(self, target):
        pass

    def try_connect(self, target, timeout=3, on_complete=None, on_failure=None):
        pass

    def disconnect(self):
        pass

    @property
    def is_connected(self):
        pass


    ### NRT Interface ###

    def init(self):
        pass

    def finish(self):
        pass


class OscUdpInterface(OscInterface):
    '''OSC over UDP server.'''

    def __init__(self, port, port_range=10):
        super().__init__(port, port_range)
        self._server = None
        self._server_thread = None
        self._running = False
        self.proto = 'udp'

    @property
    def socket(self):
        return self._server.socket if self._server else None

    def start(self):
        if self._running:
            return
        for i in range(self.port_range):
            try:
                self._server = oli.OSCUDPServer(
                    ('127.0.0.1', self._port), oli.UDPHandler)
                break
            except OSError as e:
                if e.errno == errno.EADDRINUSE and i < self.port_range - 1:
                    self._port += 1
                elif e.errno == errno.EADDRINUSE and i == self.port_range - 1:
                    err = OSError(
                        f'[Errno {errno.EADDRINUSE}] Port range already '
                        f'in use: {self.port}-{self.port_range - 1}')
                    err.errno = errno.EADDRINUSE
                    raise err from e
                else:
                    raise
        self._server_thread = threading.Thread(
            target=self._server.serve_forever,
            name=f'{type(self).__name__} port {self._port}')
        self._server_thread.daemon = True
        self._server_thread.start()
        self._running = True
        _libsc3.main._atexitq.add(
            _libsc3.main._atexitprio.NETWORKING + 1, self.stop)

    def stop(self):
        if not self._running:
            return
        self._server.shutdown()
        self._server = None
        self._running = False
        self._server_thread = None
        _libsc3.main._atexitq.remove(self.stop)

    def running(self):
        return self._running

    def _send(self, msg, target):  # override
        self._server.socket.sendto(msg.dgram, target)


class OscTcpInterface(OscInterface):
    '''
    OSC client over TCP. OscTcpInterface instances aren't reusable.
    '''

    def __init__(self, port, port_range=10):
        super().__init__(port, port_range)
        self._tcp_thread = None
        self._run_thread = False
        self._is_connected = False
        self.proto = 'tcp'

    def bind(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        for i in range(self.port_range):
            try:
                self._socket.bind(('127.0.0.1', self._port))
                break
            except OSError as e:
                if e.errno == errno.EADDRINUSE and i < self.port_range - 1:
                    self._port += 1
                elif e.errno == errno.EADDRINUSE and i == self.port_range - 1:
                    err = OSError(
                        f'[Errno {errno.EADDRINUSE}] Port range already '
                        f'in use: {self.port}-{self.port_range - 1}')
                    err.errno = errno.EADDRINUSE
                    raise err from e
                else:
                    raise

    def connect(self, target):
        self._socket.connect(target)  # Exception on failure.
        self._is_connected = True  # Not thread safe.
        self._tcp_thread = threading.Thread(
            target=self._tcp_run, name=str(self))
        self._tcp_thread.daemon = True
        self._tcp_thread.start()
        _libsc3.main._atexitq.add(
            _libsc3.main._atexitprio.NETWORKING, self.disconnect)

    def _tcp_run(self):
        self._run_thread = True
        while self._run_thread:
            try:
                data = self._socket.recv(4)
                if len(data) == 0:
                    self._is_connected = False
                    break
                size = struct.unpack('>i', data)[0]
                data = self._socket.recv(size)
                if len(data) == 0:
                    self._is_connected = False
                    break
                packet = oli.OscPacket(data)
                for timed_msg in packet.messages:
                    msg = [timed_msg.message.address, *timed_msg.message.params]
                    _libsc3.main._osc_interface.recv(
                        self._socket.getpeername(), timed_msg.time, *msg)
            except OSError as e:
                if self._run_thread:  # Log for not intentional disconnects.
                    _logger.error(f'{str(self)}: {str(e)}')
                self._is_connected = False
                break

    def try_connect(self, target, timeout=3, on_complete=None, on_failure=None):
        def tcp_connect_func():
            dt = 0.2
            attempts = int(timeout / dt)
            for i in range(attempts):
                try:
                    self.connect(target)
                    fn.value(on_complete, self)
                    return
                except ConnectionRefusedError:
                    yield dt
                except OSError as e:
                    if e.errno == errno.EADDRNOTAVAIL:  # Happens when trying to register to the existing server after a boot() fail.
                        yield dt
                    else:
                        raise
            _logger.warning(f"{str(self)} couldn't establish connection")
            fn.value(on_failure, self)

        stm.Routine.run(tcp_connect_func, clk.AppClock)

    def disconnect(self):
        self._run_thread = False
        self._tcp_thread = None
        self._socket.shutdown(socket.SHUT_RDWR)
        self._is_connected = False  # Is sync.
        self._socket.close()  # OSError if underlying error.
        _libsc3.main._atexitq.remove(self.disconnect)

    @property
    def is_connected(self):
        return self._is_connected

    def _send(self, msg, _=None):  # override
        self._socket.send(msg.size.to_bytes(4, 'big'))
        self._socket.send(msg.dgram)


class OscNrtInterface(OscInterface):
    def __init__(self):
        self._port = None
        self._port_range = None
        self._socket = None
        self._recv_functions = None
        self.proto = None

    def init(self):
        self._osc_score = OscScore()

    def finish(self):  # close?
        self._osc_score.finish()

    def add_recv_func(self, func):  # override
        pass  # Exception?

    def remove_recv_func(self, func):  # override
        pass  # Exception?

    def recv(self, addr, time, *msg):  # override
        pass  # Exception?

    def send_msg(self, target, *args):  # override
        # In NRT all messages are bundles at current time.
        self.send_bundle(target, 0.0, list(args))

    def send_bundle(self, target, time, *elements):  # override
        # Time is calculated in _build_bundle to support
        # sub-bundles relative time by calling _get_timetag().
        self._osc_score.add([time, *elements])

    @staticmethod
    def _get_timetag(time):  # override
        # Changes in this method must be synced with OscScore._get_logical_time.
        if time is None or time < 0.0:
            time = 0.0  # IMMEDIATELY is not needed in nrt.
        # In NRT bundle's time generated outside a routine is
        # always absolute time (from zero as reference time).
        if _libsc3.main.current_tt is not _libsc3.main.main_tt:
            time += _libsc3.main.current_tt.seconds
        return int(time * clk.SystemClock._SECONDS_TO_OSC)

    def _send(self, msg, target):
        pass


class OscScore():
    class _Entry():
        def __init__(self, bndl, msg):
            self.bndl = bndl
            self.msg = msg
            # *** BUG: necesita definir __eq__ y __hash__ en base a estos
            # *** BUG: parámetros para que sea único para remove, el problema
            # *** BUG: son las listas de los bndls.

    def __init__(self):
        self._scoreq = tsq.TaskQueue()
        self._lst_score = []
        self._raw_score = bytearray()
        self._finished = False
        self.add([0.0, ["/g_new", 1, 0, 0]])  # Root node.

    @property
    def list(self):
        return self._lst_score[:]

    @property
    def raw(self):
        return self._raw_score[:]

    @property
    def duration(self):
        return self._scoreq.peek(False)[0] * clk.SystemClock._OSC_TO_SECONDS

    def add(self, bndl):
        if self._finished:
            raise Exception('already finished score')
        msg = _libsc3.main._osc_interface._build_bundle(bndl)  # Raises Exception.
        msg = msg.size.to_bytes(4, 'big') + msg.dgram
        bndl = self._process_bndl_time(bndl)
        self._scoreq.add(bndl[0], type(self)._Entry(bndl, msg))

    def _process_bndl_time(self, bndl):
        # Process time in seconds to store in the score and
        # support sub-bundles relative time like _build_bundle.
        # _check_subtime is done by _build_bundle before calling this method.
        for i, element in enumerate(bndl[1:], 1):
            if isinstance(element[0], (int, float, type(None))):
                bndl[i] = self._process_bndl_time(element)
            elif not isinstance(element[0], str):
                raise ValueError(
                    'elements within bundles must be valid '
                    f'OSC messages or bundles: {element}')
        bndl = bndl[:]
        bndl[0] = self._get_logical_time(bndl[0])
        return bndl

    def _get_logical_time(self, time):
        # Same as OscNrtInterface._get_timetag but in logical time.
        # Changes in this method must be synced with it, or refactored.
        if time is None or time < 0.0:
            time = 0.0
        if _libsc3.main.current_tt is not _libsc3.main.main_tt:
            time += _libsc3.main.current_tt.seconds
        return time

    def finish(self, tail=0.0):
        if self._finished:
            return
        # This method uses the logical time of its call as the others
        # when called from a routine but when called from ouside it
        # needs to undo the check of _get_timetag and _get_logical_time.
        # Those methods and this one would need refactoring all at once.
        if _libsc3.main.current_tt is _libsc3.main.main_tt:
            tail += _libsc3.main.current_tt.seconds
        self.add([tail, ['/c_set', 0, 0]])  # Dummy cmd.
        for _, entry in self._scoreq:
            self._lst_score.append(entry.bndl)
            self._raw_score.extend(entry.msg)
        self._finished = True

    def write(self, path):
        if not self._finished:
            self.finish(self.duration)
        with open(path, 'wb') as file:
            file.write(self._raw_score)

    def render(self, path=None, input_file=None, server=None):
        # This method blocks until cmd finish and returns its exit code.
        osc_file = plf.Platform.tmp_dir
        osc_file /= 'SC_' + time.strftime('%Y%m%d_%H%M%S') + '.osc'
        self.write(osc_file)

        server = srv.Server.default if server is None else server
        cmd = [server.options.program]
        cmd.extend(
            server.options.options_list(None, osc_file, input_file, path))

        self._render_proc = subprocess.Popen(
            cmd,
            stdin=subprocess.DEVNULL,
            stdout=sys.stdout,
            stderr=sys.stderr,
            bufsize=1,
            universal_newlines=True)

        self._render_proc.wait()
        osc_file.unlink()
        return self._render_proc.poll()

    def __str__(self):
        return pprint.pformat(self._lst_score)



"""Filter.sc"""

from .. import ugen as ugn
from .. import _graphparam as gpp

# Only for VarLag
from .. import envelope as evp
from . import oscillators as ocl
from . import envgen as evg


class Filter(ugn.UGen, ugn.PureUGen):
    def _check_inputs(self):  # override
        return self._check_sr_as_first_input()


class Resonz(Filter):
    @classmethod
    def ar(cls, input=0.0, freq=440.0, bwr=1.0):
        return cls._multi_new('audio', input, freq, bwr)

    @classmethod
    def kr(cls, input=0.0, freq=440.0, bwr=1.0):
        return cls._multi_new('control', input, freq, bwr)


class OnePole(Filter):
    @classmethod
    def ar(cls, input=0.0, coef=0.5):
        return cls._multi_new('audio', input, coef)

    @classmethod
    def kr(cls, input=0.0, coef=0.5):
        return cls._multi_new('control', input, coef)


class OneZero(OnePole):
    pass


class TwoPole(Filter):
    @classmethod
    def ar(cls, input=0.0, freq=440.0, radius=0.8):
        return cls._multi_new('audio', input, freq, radius)

    @classmethod
    def kr(cls, input=0.0, freq=440.0, radius=0.8):
        return cls._multi_new('control', input, freq, radius)


class TwoZero(TwoPole):
    pass


class APF(TwoPole):
    pass


class Integrator(Filter):
    @classmethod
    def ar(cls, input=0.0, coef=1.0):
        return cls._multi_new('audio', input, coef)

    @classmethod
    def kr(cls, input=0.0, coef=1.0):
        return cls._multi_new('control', input, coef)


class Decay(Filter):
    @classmethod
    def ar(cls, input=0.0, decay_time=1.0):
        return cls._multi_new('audio', input, decay_time)

    @classmethod
    def kr(cls, input=0.0, decay_time=1.0):
        return cls._multi_new('control', input, decay_time)


class Decay2(Filter):
    @classmethod
    def ar(cls, input=0.0, attack_time=0.01, decay_time=1.0):
        return cls._multi_new('audio', input, attack_time, decay_time)

    @classmethod
    def kr(cls, input=0.0, attack_time=0.01, decay_time=1.0):
        return cls._multi_new('control', input, attack_time, decay_time)


class Lag(Filter):
    @classmethod
    def ar(cls, input=0.0, lag_time=0.1):
        if gpp.ugen_param(input)._as_ugen_rate() == 'scalar' or lag_time == 0:
            return input
        else:
            return cls._multi_new('audio', input, lag_time)

    @classmethod
    def kr(cls, input=0.0, lag_time=0.1):
        if gpp.ugen_param(input)._as_ugen_rate() == 'scalar' or lag_time == 0:
            return input
        else:
            return cls._multi_new('control', input, lag_time)


class Lag2(Lag):
    pass


class Lag3(Lag):
    pass


class Ramp(Lag):
    pass


class LagUD(Filter):
    @classmethod
    def ar(cls, input=0.0, lag_time_up=0.1, lag_time_down=0.1):
        if gpp.ugen_para(input)._as_ugen_rate() == 'scalar':
            return input
        else:
            return cls._multi_new('audio', input, lat_time_up, lag_time_down)

    @classmethod
    def kr(cls, input=0.0, lag_time_up=0.1, lag_time_down=0.1):
        if gpp.ugen_para(input)._as_ugen_rate() == 'scalar':
            return input
        else:
            return cls._multi_new('control', input, lat_time_up, lag_time_down)


class Lag2UD(LagUD):
    pass


class Lag3UD(LagUD):
    pass


class VarLag(Filter):
    @classmethod
    def ar(cls, input=0.0, time=0.1, curvature=0, warp=5, start=None):
        if gpp.ugen_param(input)._as_ugen_rate() == 'scalar' or time == 0:
            return input
        else:
            return cls._multi_new('audio', input, time, curvature, warp, start)

    @classmethod
    def kr(cls, input=0.0, time=0.1, curvature=0, warp=5, start=None):
        if gpp.ugen_param(input)._as_ugen_rate() == 'scalar' or time == 0:
            return input
        else:
            return cls._multi_new(
                'control', input, time, curvature, warp, start)

    @classmethod
    def _new1(cls, rate, input, time, curvature, warp, start):   # *** BUG: TEST.
        # // FIXME: Implement 'curve' input on VLag ugen instead of using EnvGen.
        # // Then 'exp' warp should probably behave as Lag ugen.
        # NOTE: this provisory implementation is low level like in sclang.
        selector = 'ar' if rate == 'audio' else 'kr'
        if start is None:
            start = input
        try:
            curve = evp.Env._SHAPE_NAMES[warp]
        except KeyError:
            curve = warp
        if curve != 1:
            env = evp.Env([start, input], [time], warp).envgen_format()
            env = list(env[0])
            env[6] = curve
            env[7] = curvature
            env = [tuple(env)]
            trig = getattr(Changed, selector)(input) +\
                   getattr(ocl.Impulse, selector)(0)
            if gpp.ugen_para(time)._as_ugen_rate() != 'scalar':
                trig += Changed.kr(time)
            return getattr(evg.EnvGen, selector)(env, trig)
        else:
            obj = cls._create_ugen_object(rate)
            obj._add_to_synth()
            return obj._init_ugen(input, time, start)


class LeakDC(Filter):
    @classmethod
    def ar(cls, input=0.0, coef=0.995):
        return cls._multi_new('audio', input, coef)

    @classmethod
    def kr(cls, input=0.0, coef=0.995):
        return cls._multi_new('control', input, coef)


class RLPF(Filter):
    @classmethod
    def ar(cls, input=0.0, freq=440.0, rq=1.0):
        return cls._multi_new('audio', input, freq, rq)

    @classmethod
    def kr(cls, input=0.0, freq=440.0, rq=1.0):
        return cls._multi_new('control', input, freq, rq)


class RHPF(RLPF):
    pass


class LPF(Filter):
    @classmethod
    def ar(cls, input=0.0, freq=440.0):
        return cls._multi_new('audio', input, freq)

    @classmethod
    def kr(cls, input=0.0, freq=440.0):
        return cls._multi_new('control', input, freq)


class HPF(LPF):
    pass


class BPF(Filter):
    @classmethod
    def ar(cls, input=0.0, freq=440.0, rq=1.0):
        return cls._multi_new('audio', input, freq, rq)

    @classmethod
    def kr(cls, input=0.0, freq=440.0, rq=1.0):
        return cls._multi_new('control', input, freq, rq)


class BRF(BPF):
    pass


class MidEQ(Filter):
    @classmethod
    def ar(cls, input=0.0, freq=440.0, rq=1.0, db=0.0):
        return cls._multi_new('audio', input, freq, rq, db)

    @classmethod
    def kr(cls, input=0.0, freq=440.0, rq=1.0, db=0.0):
        return cls._multi_new('control', input, freq, rq, db)


class LPZ1(Filter):
    @classmethod
    def ar(cls, input=0.0):
        return cls._multi_new('audio', input)

    @classmethod
    def kr(cls, input=0.0):
        return cls._multi_new('control', input)


class HPZ1(LPZ1):
    pass


class Slope(Filter):
    @classmethod
    def ar(cls, input=0.0):
        return cls._multi_new('audio', input)

    @classmethod
    def kr(cls, input=0.0):
        return cls._multi_new('control', input)


class Changed(Filter):  # is pseudo really
    @classmethod
    def ar(cls, input, threshold=0):
        return HPZ1.ar(input).abs() > threshold

    @classmethod
    def kr(cls, input, threshold=0):
        return HPZ1.kr(input).abs() > threshold


class LPZ2(Filter):
    @classmethod
    def ar(cls, input=0.0):
        return cls._multi_new('audio', input)

    @classmethod
    def kr(cls, input=0.0):
        return cls._multi_new('control', input)


class HPZ2(LPZ2):
    pass


class BPZ2(LPZ2):
    pass


class BRZ2(LPZ2):
    pass


class Median(Filter):
    @classmethod
    def ar(cls, length=3, input=0.0):
        return cls._multi_new('audio', length, input)

    @classmethod
    def kr(cls, length=3, input=0.0):
        return cls._multi_new('control', length, input)

    def _check_inputs(self):
        if self.rate == 'audio'\
        and gpp.ugen_para(self.inputs[1])._as_ugen_rate() != 'audio':
            return 'input was not audio rate'
        else:
            return self._check_valid_inputs()


# Lost UGen AvgAbsAmp.


class Slew(Filter):
    @classmethod
    def ar(cls, input=0.0, up=1.0, dn=1.0):
        if gpp.ugen_param(input)._as_ugen_rate() == 'scalar':
            return input
        else:
            return cls._multi_new('audio', input, up, dn)

    @classmethod
    def kr(cls, input=0.0, up=1.0, dn=1.0):
        if gpp.ugen_param(input)._as_ugen_rate() == 'scalar':
            return input
        else:
            return cls._multi_new('control', input, up, dn)


# Lost UGen RLPF4.


class FOS(Filter):
    @classmethod
    def ar(cls, input=0.0, a0=0.0, a1=0.0, b1=0.0):
        return cls._multi_new('audio', input, a0, a1, b1)

    @classmethod
    def kr(cls, input=0.0, a0=0.0, a1=0.0, b1=0.0):
        return cls._multi_new('control', input, a0, a1, b1)


class SOS(Filter):
    @classmethod
    def ar(cls, input=0.0, a0=0.0, a1=0.0, a2=0.0, b1=0.0, b2=0.0):
        return cls._multi_new(
            'audio', input, a0, a1, a2, b1, b2)

    @classmethod
    def kr(cls, input=0.0, a0=0.0, a1=0.0, a2=0.0, b1=0.0, b2=0.0):
        return cls._multi_new(
            'control', input, a0, a1, a2, b1, b2)


class Ringz(Filter):
    @classmethod
    def ar(cls, input=0.0, freq=440.0, decay_time=1.0):
        return cls._multi_new('audio', input, freq, decay_time)

    @classmethod
    def kr(cls, input=0.0, freq=440.0, decay_time=1.0):
        return cls._multi_new('control', input, freq, decay_time)


class Formlet(Filter):
    @classmethod
    def ar(cls, input=0.0, freq=440.0, attack_time=1.0, decay_time=1.0):
        return cls._multi_new('audio', input, freq, attack_time, decay_time)

    @classmethod
    def kr(cls, input=0.0, freq=440.0, attack_time=1.0, decay_time=1.0):
        return cls._multi_new('control', input, freq, attack_time, decay_time)


# // the doneAction arg lets you cause the EnvGen to stop or end the
# // synth without having to use a PauseSelfWhenDone or FreeSelfWhenDone ugen.
# // It is more efficient to use a doneAction.
# // doneAction = 0   do nothing when the envelope has ended.
# // doneAction = 1   pause the synth running, it is still resident.
# // doneAction = 2   remove the synth and deallocate it.
# // doneAction = 3   remove and deallocate both this synth and the preceeding node.
# // doneAction = 4   remove and deallocate both this synth and the following node.
# // doneAction = 5   remove and deallocate this synth and free all children in the preceeding group (if it is a group).
# // doneAction = 6   remove and deallocate this synth and free all children in the following group (if it is a group).
class DetectSilence(Filter):
    @classmethod
    def ar(cls, input=0.0, amp=0.0001, time=0.1, done_action=0):
        return cls._multi_new('audio', input, amp, time, done_action)

    @classmethod
    def kr(cls, input=0.0, amp=0.0001, time=0.1, done_action=0):
        return cls._multi_new('control', input, amp, time, done_action)

    def _optimize_graph(self):
        pass


# Lost UGen FlagNaN.



"""Ppatmod.sc"""

from ...base import stream as stm
from .. import pattern as ptt


class Pvalue(ptt.Pattern):
    # This pattern is for special cases where common
    # values aren't or can't be embeded as streams.
    def __init__(self, value):
        self.value = value

    def __embed__(self, inval):
        return (yield from stm.stream(self.value))

    # storeArgs


class Plazy(ptt.Pattern):
    def __init__(self, func):
        self.func = func

    def __embed__(self, inval):
        return (yield from stm.embed(self.func(inval), inval))

    # storeArgs



"""From Patterns.sc"""

import inspect

from ...base import stream as stm
from ...base import builtins as bi
from .. import pattern as ptt


### Function patterns ###


class Pfunc(ptt.Pattern):
    def __init__(self, next_func, reset_func=None, data=None):
        self.next_func = next_func
        self.reset_func = reset_func
        self.data = data

    def __stream__(self):
        return stm.FunctionStream(self.next_func, self.reset_func, self.data)

    # storeArgs


class Prout(ptt.Pattern):
    def __init__(self, func):
        self.func = func
        self._func_has_inval = (  # See note in TimeThread.__init__. Sync code.
            len(inspect.signature(self.func).parameters) > 0)
        self._func_isgenfunc = inspect.isgeneratorfunction(self.func)

    def __stream__(self):
        return stm.Routine(self.func)

    def __embed__(self, inval):
        if self._func_isgenfunc:
            if self._func_has_inval:
                iterator = self.func(inval)
            else:
                iterator = self.func()
            try:
                yield next(iterator)
                while True:
                    yield iterator.send(inval)
            except StopIteration as e:
                return e.value  # Contains generator function's return value.
        else:
            if self._func_has_inval:
                return self.func(inval)
            else:
                return self.func()

    # storeArgs


class Pfuncn(ptt.Pattern):
    def __init__(self, func, repeats=1):
        self.func = func
        self._func_has_inval = (  # See note in TimeThread.__init__. Sync code.
            len(inspect.signature(self.func).parameters) > 0)
        self.repeats = repeats

    def __embed__(self, inval):
        for i in bi.counter(self.repeats):
            if self._func_has_inval:
                inval = yield self.func(inval)
            else:
                inval = yield self.func()
        return inval

    # storeArgs


# BUG: ver su utilidad, qué diferencia hay
# con usar un operador enario directamente?
# class PdegreeToKey(Pnarop):
#     ...



"""
Builtin UGen classes package.
"""

import sys as _sys

from ...base import _hooks as hks


installed_ugens = dict()


def install_ugen(ugen):
    '''
    Makes a single ugen available to sc3.synth.ugens.installed_ugens (required
    by SynthDesc for the ugen to be valid).
    '''
    entry = {ugen.__name__: ugen}
    _sys.modules[__name__].__dict__.update(entry)
    installed_ugens.update(entry)


def install_ugens_module(name, package=None):
    '''
    At the end of an external ugens module call install_ugens_module(__name__)
    to make all classes available in sc3.synth.ugens.installed_ugens.
    Parameters name and package are the arguments of importlib.import_module().
    '''
    mapping = hks.import_all_module(name, package, bind=__name__)
    installed_ugens.update(mapping)


def install_ugens_package(path, name):
    '''
    Call this function in the __init__.py file of an ugens package with
    __path__ and __name__ as arguments to make all classes available in
    sc3.synth.ugens.installed_ugens. Parameters path and name are used as
    arguments for pkgutil.walk_packages().
    '''
    mapping = hks.import_all_package(path, name, bind=__name__)
    installed_ugens.update(mapping)


install_ugens_module('sc3.synth.ugen')  # Not really user classes.
install_ugens_package(__path__, __name__)

__all__ = list(installed_ugens.keys())



"""Pan.sc"""

from .. import ugen as ugn


class Pan2(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, input, pos=0.0, level=1.0):
        return cls._multi_new('audio', input, pos, level)

    @classmethod
    def kr(cls, input, pos=0.0, level=1.0):
        return cls._multi_new('control', input, pos, level)

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList([
            ugn.OutputProxy.new(self.rate, self, 0),
            ugn.OutputProxy.new(self.rate, self, 1)
        ])
        return self._channels # NOTE: RECORDAR: las ugens retornan self en _init_ugen que es método de interfaz, pero las output ugens retornan self._channels (o _init_outputs que retorna self._channels)

    def _check_inputs(self):  # override
        return self._check_n_inputs(1)


class LinPan2(Pan2):
    pass


class Pan4(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, input, xpos=0.0, ypos=0.0, level=1.0):
        return cls._multi_new('audio', input, xpos, ypos, level)

    @classmethod
    def kr(cls, input, xpos=0.0, ypos=0.0, level=1.0):
        return cls._multi_new('control', input, xpos, ypos, level)

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList([
            ugn.OutputProxy.new(self.rate, self, 0),
            ugn.OutputProxy.new(self.rate, self, 1),
            ugn.OutputProxy.new(self.rate, self, 2),
            ugn.OutputProxy.new(self.rate, self, 3)
        ])
        return self._channels

    def _check_inputs(self):  # override
        return self._check_n_inputs(1)


class Balance2(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, left, right, pos=0.0, level=1.0):
        return cls._multi_new('audio', left, right, pos, level)

    @classmethod
    def kr(cls, left, right, pos=0.0, level=1.0):
        return cls._multi_new('control', left, right, pos, level)

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList([
            ugn.OutputProxy.new(self.rate, self, 0),
            ugn.OutputProxy.new(self.rate, self, 1)
        ])
        return self._channels

    def _check_inputs(self):  # override
        return self._check_n_inputs(2)


class Rotate2(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, x, y, pos=0.0):
        return cls._multi_new('audio', x, y, pos)

    @classmethod
    def kr(cls, x, y, pos=0.0):
        return cls._multi_new('control', x, y, pos)

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList([
            ugn.OutputProxy.new(self.rate, self, 0),
            ugn.OutputProxy.new(self.rate, self, 1)
        ])
        return self._channels

    def _check_inputs(self):  # override
        return self._check_n_inputs(2)


class PanB(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, input, azimuth=0, elevation=0, gain=1):
        return cls._multi_new('audio', input, azimuth, elevation, gain)

    @classmethod
    def kr(cls, input, azimuth=0, elevation=0, gain=1):
        return cls._multi_new('control', input, azimuth, elevation, gain)

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList([
            ugn.OutputProxy.new(self.rate, self, 0),
            ugn.OutputProxy.new(self.rate, self, 1),
            ugn.OutputProxy.new(self.rate, self, 2),
            ugn.OutputProxy.new(self.rate, self, 3)
        ])
        return self._channels

    def _check_inputs(self):  # override
        return self._check_n_inputs(1)


class PanB2(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, input, azimuth=0, gain=1):
        return cls._multi_new('audio', input, azimuth, gain)

    @classmethod
    def kr(cls, input, azimuth=0, gain=1):
        return cls._multi_new('control', input, azimuth, gain)

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList([
            ugn.OutputProxy.new(self.rate, self, 0),
            ugn.OutputProxy.new(self.rate, self, 1),
            ugn.OutputProxy.new(self.rate, self, 2)
        ])
        return self._channels

    def _check_inputs(self):  # override
        return self._check_n_inputs(1)


class BiPanB2(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, input_a, input_b, azimuth=0, gain=1):
        return cls._multi_new('audio', input_a, input_b, azimuth, gain)

    @classmethod
    def kr(cls, input_a, input_b, azimuth=0, gain=1):
        return cls._multi_new('control', input_a, input_b, azimuth, gain)

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList([
            ugn.OutputProxy.new(self.rate, self, 0),
            ugn.OutputProxy.new(self.rate, self, 1),
            ugn.OutputProxy.new(self.rate, self, 2)
        ])
        return self._channels

    def _check_inputs(self):  # override
        return self._check_n_inputs(2)


class DecodeB2(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels, w, x, y, orientation=0.5):
        return cls._multi_new('audio', num_channels, w, x, y, orientation)  # *** BUG: en sclang? asigns orientation = 0.5 here again.

    @classmethod
    def kr(cls, num_channels, w, x, y, orientation=0.5):
        return cls._multi_new('control', num_channels, w, x, y, orientation)  # *** BUG: en sclang? asigns orientation = 0.5 here again.

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList([
            ugn.OutputProxy.new(self.rate, self, i)
            for i in range(num_channels)
        ])
        return self._channels

    def _check_inputs(self):  # override
        return self._check_n_inputs(3)


class PanAz(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels, input, pos=0.0, level=1.0,
           width=2.0, orientation=0.5):
        return cls._multi_new('audio', num_channels, input, pos, level,
                              width, orientation)

    @classmethod
    def kr(cls, num_channels, input, pos=0.0, level=1.0,
           width=2.0, orientation=0.5):
        return cls._multi_new('control', num_channels, input, pos, level,
                              width, orientation)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList([
            ugn.OutputProxy.new(self.rate, self, i)
            for i in range(num_channels)
        ])
        return self._channels

    def _check_inputs(self):  # override
        return self._check_n_inputs(1)


class XFade2(ugn.UGen):
    # // Equal power two channel cross fade.
    @classmethod
    def ar(cls, input_a, input_b=0.0, pan=0.0, level=1.0):
        return cls._multi_new('audio', input_a, input_b, pan, level)

    @classmethod
    def kr(cls, input_a, input_b=0.0, pan=0.0, level=1.0):
        return cls._multi_new('control', input_a, input_b, pan, level)

    def _check_inputs(self):  # override
        return self._check_n_inputs(2)


class LinXFade2(ugn.UGen):
    # // Linear two channel cross fade.
    @classmethod
    def ar(cls, input_a, input_b=0.0, pan=0.0, level=1.0):
        return cls._multi_new('audio', input_a, input_b, pan) * level

    @classmethod
    def kr(cls, input_a, input_b=0.0, pan=0.0, level=1.0):
        return cls._multi_new('control', input_a, input_b, pan) * level

    def _check_inputs(self):  # override
        return self._check_n_inputs(2)



"""NetAddr.sc"""

import ipaddress
import socket

from . import stream as stm
from . import main as _libsc3
from . import builtins as bi
from . import responsedefs as rdf
from . import _oscinterface as osci


__all__ = ['NetAddr', 'BundleNetAddr']


class NetAddr():
    # use_doubles = False
    # broadcast_flag = False

    # From Wikipedia: "The field size sets a theoretical limit of 65,535 bytes
    # (8 byte header + 65,527 bytes of data) for a UDP datagram. However the
    # actual limit for the data length, which is imposed by the underlying
    # IPv4 protocol, is 65,507 bytes (65,535 − 8 byte UDP header − 20 byte
    # IP header)".
    _MAX_UDP_DGRAM_SIZE = 65507
    _SYNC_BNDL_DGRAM_SIZE = 36  # Size of bundle(latency, ['/sync', id]) dgram.

    def __init__(self, hostname, port):
        if hostname is None:
            self._addr = 0  # Server:in_process
        else:
            self._addr = int(ipaddress.IPv4Address(hostname))  # Exception if invalid.
        self._hostname = hostname
        self._port = port
        self._target = (hostname, port)
        self._tcp_interface = None

    @property
    def hostname(self):
        return self._hostname

    @property
    def addr(self):
        return self._addr

    @property
    def port(self):
        return self._port

    @property
    def proto(self):
        return self._osc_interface.proto

    @property
    def is_local(self):
        return self.match_lang_ip(self._hostname)

    @property
    def local_end_point(self):  # bind_addr
        if self._osc_interface.socket:
            return type(self)(*self._osc_interface.socket.getsockname())

    @property
    def _osc_interface(self):
        return self._tcp_interface or _libsc3.main._osc_interface

    @staticmethod
    def lang_port():
        return _libsc3.main._osc_interface.port

    @staticmethod
    def match_lang_ip(ipstring):
        # if ipaddress.IPv4Address(self._addr).is_loopback:
        if ipstring == '127.0.0.1':
            return True
        addr_info = socket.getaddrinfo(socket.gethostname(), None)
        for item in addr_info:
            if item[4][0] == ipstring:
                return True
        return False


    ### TCP Connections ###

    def connect(self, on_complete=None, on_failure=None,
                local_port=None, port_range=10, timeout=3):
        # Async.
        if self.is_connected:
            self.disconnect()
        self._tcp_interface = osci.OscTcpInterface(
            local_port or self.lang_port() + 1, port_range)
        self._tcp_interface.bind()
        self._tcp_interface.try_connect(
            self._target, timeout, on_complete, on_failure)

    def disconnect(self):
        # Sync.
        self._tcp_interface.disconnect()
        self._tcp_interface = None

    @property
    def is_connected(self):
        return self._tcp_interface and self._tcp_interface.is_connected


    def has_bundle(self):
        return False

    # def send_raw(self, raw_bytes):
    #     # send a raw message without timestamp to the addr.
    #     self._osc_interface.send_raw(self._target, raw_bytes)

    def send_msg(self, *args):
        self._osc_interface.send_msg(self._target, *args)

    def send_bundle(self, time, *elements):
        self._osc_interface.send_bundle(self._target, time, *elements)

    def send_clumped_bundles(self, time, *elements):
        if self._calc_bndl_dgram_size(elements) > self._MAX_UDP_DGRAM_SIZE:
            for item in self._clump_bundle(elements):
                if time is not None:
                    time += 1e-9  # One nanosecond later each.
                self.send_bundle(time, *item)
        else:
            self.send_bundle(time, *elements)

    def send_status_msg(self):
        self._osc_interface.send_msg(self._target, '/status')

    def recover(self):
        return self

    def sync(self, condition=None, latency=None, elements=None):
        condition = condition or stm.Condition()
        if elements is None:
            id = self._make_sync_responder(condition)
            self.send_bundle(latency, ['/sync', id])
            yield from condition.wait()
        else:
            sync_size = self._SYNC_BNDL_DGRAM_SIZE
            max_size = self._MAX_UDP_DGRAM_SIZE - sync_size
            if self._calc_bndl_dgram_size(elements) > max_size:
                for item in self._clump_bundle(elements, max_size):
                    id = self._make_sync_responder(condition)
                    item.append(['/sync', id])
                    self.send_bundle(latency, *item)
                    if latency is not None:
                        latency += 1e-9  # One nanosecond later each.
                    yield from condition.wait()
            else:
                id = self._make_sync_responder(condition)
                elements = list(elements)
                elements.append(['/sync', id])
                self.send_bundle(latency, *elements)
                yield from condition.wait()

    def _make_sync_responder(self, condition):
        id = bi.uid()

        def resp_func(msg, *_):
            if msg[1] == id:
                resp.free()
                condition.test = True
                condition.signal()

        resp = rdf.OscFunc(resp_func, '/synced', self)
        return id

    def _clump_bundle(self, elements, size=8192):
        elist = []
        for e in elements:
            if isinstance(e[0], str):
                elist.append((self._calc_msg_dgram_size(e), e))
            elif isinstance(e[0], (int, float)):  # bundle
                elist.append((self._calc_bndl_dgram_size(e[1:]), e))
            else:
                raise ValueError(
                    'elements within bundles must be valid OSC '
                    f'messages or bundles, received: {e}')
        res = []
        clump = []
        acc_size = 16  # Bundle prefix + Timetag bytes.
        for s, e in elist:
            if acc_size + s >= size:
                res.append(clump)
                clump = []
                acc_size = 16  # Bundle prefix + Timetag bytes.
            acc_size += s
            clump.append(e)
        if clump:
            res.append(clump)
        return res

    def _calc_bndl_dgram_size(self, elements):
        # Argument elements is the content without Timetag: [[], [], ...].
        res = 16  # Bundle prefix + Timetag bytes.
        for e in elements:
            res += 4  # Element size bytes.
            if isinstance(e[0], str):  # message
                res += self._calc_msg_dgram_size(e)
            elif isinstance(e[0], (int, float)):  # bundle
                res += self._calc_bndl_dgram_size(e[1:])
            else:
                raise ValueError(
                    'elements within bundles must be valid OSC '
                    f'messages or bundles, received: {e}')
        return res

    def _calc_msg_dgram_size(self, msg):
        res = self._strpad4(len(bytes(msg[0], 'ascii')))  # Address.
        res += self._strpad4(len(msg[1:]) + 1)  # Type tag string.
        for val in msg[1:]:
            if isinstance(val, str):
                res += self._strpad4(len(val))
            elif isinstance(val, list):
                # sc3 don't send osc arrays, they are msgs converted to blobs.
                res += self._calc_msg_dgram_size(val) + 4  # Blob size bytes.
            else:
                res += 4  # Everything else (sent by sc3, no doubles).
        return res

    @staticmethod
    def _strpad4(n):
        # Pad to 4 or add null if mod is zero.
        return n + 4 - (n & 3)

    def __eq__(self, other):
        if type(self) == type(other):
            return self._target == other._target
        else:
            return False

    def __hash__(self):
        return hash((type(self), hash(self._target)))

    def __repr__(self):
        return f"{type(self).__name__}('{self.hostname}', {self.port})"


class BundleNetAddr(NetAddr):
    # Important difference: This class is a context manager. forkIfNeeded
    # can't be implemented, addr.sync() uses it in sclang. Here sync calls are
    # handled different doing yield from directly within the with statement.
    # I don't see a significant difference with sending all sync messages
    # together at the end.

    class _SYNC_FLAG(): pass

    def __init__(self, target, arg_list=None, send=True):
        if isinstance(target, NetAddr):
            self._save_addr = addr
            self._server = None
        else:  # Assumes it's a Server to avoid the import.
            self._save_addr = target.addr
            self._server = target
        super().__init__(self._save_addr._hostname, self._save_addr._port)
        self._bundle = arg_list or []
        self._send = send
        self._last_sync = -1

    def has_bundle(self):
        return True

    # def send_raw(self, rawlst):

    def send_msg(self, *args):
        self._bundle.append(list(args))

    def send_bundle(self, time, *elements):
        self._bundle.extend(list(elements))  # Discard time.

    def send_clumped_bundles(self, time, *elements):
        self._bundle.extend(list(elements))  # Discard time.

    def send_status_msg(self):
        pass  # // Ignore status message.

    def recover(self):
        return self._save_addr

    def sync(self, condition=None, latency=None, elements=None):
        if self._send:
            self._send_last_bundle()
            yield from self._save_addr.sync(None, latency, elements)
        self._last_sync = len(self._bundle)
        self._bundle.append([self._SYNC_FLAG, latency, elements])

    def _send_last_bundle(self):
        time = self._server.latency if self._server else None
        bundle = self._bundle[self._last_sync+1:]
        if bundle:
            self._save_addr.send_clumped_bundles(time, *bundle)

    def _split_bundles(self, time):
        res = []
        curr = [time]
        for item in self._bundle:
            if item[0] is self._SYNC_FLAG:
                if item[2] is not None:
                    curr.extend(item[2])
                res.append(curr)
                curr = [item[1]]
            else:
                curr.append(item)
        res.append(curr)
        return res

    def get_bundle(self, time=None):
        if self._last_sync == -1:
            return [time, *self._bundle]
        else:
            return self._split_bundles(time)

    def __enter__(self):
        if self._server:
            self._server._addr = self
        # return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._server:
            self._server._addr = self._save_addr
        if exc_type is None and self._send:
            self._send_last_bundle()




from . buffer import *
from . bus import *
from . envelope import *
from . node import *
from . server import *
from . synthdef import *
from . synthdesc import *
from . spec import *
# from . ugen import *  # Done by ugens.
from . ugens import *
from . systemdefs import *



"""Node.sc"""

import logging

from ..base import utils as utl
from ..base import responsedefs as rdf
from ..base import functions as fn
from ..base import model as mdl
from ..base import stream as stm
from ..base import clock as clk
from . import server as srv
from . import synthdesc as sdc
from . import _graphparam as gpp


__all__ = ['Group', 'ParGroup', 'Synth']


_logger = logging.getLogger(__name__)


class Node(gpp.NodeParameter):
    """Base class for ``Group`` and ``Synth``.

    This class is not used directly but though its subclasses, ``Synth``
    and ``Group`` (or ``ParGroup``), which represent synth or group
    nodes on the server.

    Node objects which you explicitly free using the methods ``free``
    or ``release`` will have their ``group`` instance variable set to
    ``None``. However, objects which are automatically freed after a
    certain time (for instance by an ``EnvGen`` with a done action of 2)
    will not. This keeps the implementation of the classes simple and
    lightweight.

    To have the current state of a node tracked you can register it
    with an instance of ``NodeWatcher``, either by calling register on
    the ``Node`` instance or on the ``NodeWatcher`` singleton. This will
    enable two variables, ``is_playing`` and ``is_running``, which you
    can use for checking purposes.

    Many of the methods below have two versions: a regular one which
    sends its corresponding message to the server immediately, and one
    which returns the message in a list so that it can be added to a
    bundle. It is also possible to capture the messages generated by
    the regular methods using Server's automated bundling capabilities.
    """
    # TODO: :Bundled Server Messages: is not inline docstring.

    add_actions = {
        # Traditional.
        'addToHead': 0,
        'addToTail': 1,
        'addBefore': 2,
        'addAfter': 3,
        'addReplace': 4,
        # Simple.
        'head': 0,
        'tail': 1,
        'before': 2,
        'after': 3,
        'replace': 4,
        # Shortcut.
        'h': 0, 't': 1, 'b': 2, 'a': 3, 'r': 4,
        # // valid action numbers should stay the same
        0: 0, 1: 1, 2: 2, 3: 3, 4: 4
    }

    _register_all = False

    def __init__(self):
        super(gpp.NodeParameter, self).__init__(self)

    def _init_register(self, register):
        self._is_playing = None  # None (not watched/no info), True or False
        self._is_running = None  # None (not watched/no info), True or False
        if register or Node._register_all:
            self.register()

    @property
    def is_playing(self):
        return self._is_playing

    @property
    def is_running(self):
        return self._is_running

    @classmethod
    def basic_new(cls, server=None, node_id=None):
        obj = cls.__new__(cls)  # basic_new doesn't send therefore can't call __init__
        super(gpp.NodeParameter, obj).__init__(obj)
        obj.server = server or srv.Server.default
        if node_id is None:
            obj.node_id = obj.server.next_node_id()
        else:
            obj.node_id = node_id
        obj.group = None
        obj._is_playing = None  # None (not watched/no info), True or False
        obj._is_running = None  # None (not watched/no info), True or False
        return obj

    @classmethod
    def action_number_for(cls, add_action): # ='addToHead'): BUG: en sclang, semánticamente no tiene sentido el valor por defecto
        return cls.add_actions[add_action]

    def free(self, send_flag=True):
        if send_flag:
            self.server.send_msg('/n_free', self.node_id) # 11
        self.group = None

    def free_msg(self):
        return ('/n_free', self.node_id) # 11

    def run(self, flag=True):
        self.server.send_msg('/n_run', self.node_id, int(flag)) # 12

    def run_msg(self, flag=True):
        return ('/n_run', self.node_id, int(flag)) # 12

    def map(self, *args):
        bundle = self.map_msg(*args)
        if isinstance(bundle[0], str):
            self.server.send_bundle(0, bundle)
        else:
            self.server.send_bundle(0, *bundle)

    def map_msg(self, *args):
        kr_values = []
        ar_values = []
        result = []
        for control, bus in utl.gen_cclumps(args, 2):
            # No as_bus cast, Bus object is required for all
            # rate cases because the method joins /m_mapan.
            if bus.rate == 'control':
                kr_values.extend([
                    gpp.node_param(control)._as_control_input(),
                    bus.index,
                    bus.channels
                ])
            elif bus.rate == 'audio':
                ar_values.extend([
                    gpp.node_param(control)._as_control_input(),
                    bus.index,
                    bus.channels
                ])
            # // no default case, ignore others
        if len(kr_values) > 0:
            result.append(['/n_mapn', self.node_id] + kr_values)
        if len(ar_values) > 0:
            result.append(['/n_mapan', self.node_id] + ar_values)
        if len(result) < 2:
            result = utl.flatten(result)
        return result

    def mapn(self, *args):
        self.server.send_msg(
            '/n_mapn', # 48
            self.node_id,
            *gpp.node_param(args)._as_control_input()
        )

    def mapn_msg(self, *args):
        return ['/n_mapn', self.node_id]\
            + gpp.node_param(args)._as_control_input() # 48

    def set(self, *args):
        self.server.send_msg(
            '/n_set', # 15
            self.node_id,
            *gpp.node_param(args)._as_osc_arg_list()
        )

    def set_msg(self, *args):
        return ['/n_set', self.node_id]\
            + gpp.node_param(args)._as_osc_arg_list() # 15

    def setn(self, *args):
        self.server.send_msg(*self.setn_msg(*args))

    def setn_msg(self, *args):
        return ['/n_setn', self.node_id] + self._setn_msg_args(*args) # 16

    @staticmethod
    def _setn_msg_args(*args):
        nargs = []
        args = gpp.node_param(args)._as_control_input()
        for control, more_vals in utl.gen_cclumps(args, 2):
            if isinstance(more_vals, list): # BUG: ídem acá arriba, more_vals TIENE QUE SER LISTA
                nargs.extend([control, len(more_vals)] + more_vals)
            else:
                nargs.extend([control, 1, more_vals])
        return nargs

    def fill(self, cname, num_controls, value, *args):
        self.server.send_msg(
            '/n_fill', self.node_id, # 17
            cname, num_controls, value,
            *gpp.node_param(args)._as_control_input()
        )

    def fill_msg(self, cname, num_controls, value, *args):
        return ['/n_fill', self.node_id, cname, num_controls, value]\
            + gpp.node_param(args)._as_control_input() # 17

    def release(self, time=None):
        # Sends a bundle so it can be used as
        # counterpart of SynthDef__call__.
        self.server.send_bundle(0, self.release_msg(time))

    def release_msg(self, time=None):
        # // assumes a control called 'gate' in the synth
        if time is not None:
            if time <= 0:
                time = -1
            else:
                time = -(time + 1)
        else:
            time = 0
        return ['/n_set', self.node_id, 'gate', time] # 15

    def trace(self):
        self.server.send_msg('/n_trace', self.node_id) # 10

    def query(self, action=None):
        if action is None:
            def action(cmd, node_id, parent, prev, next,
                       is_group, head=None, tail=None):
                group = is_group == 1
                node_type = 'Group' if group else 'Synth'
                msg = (f'{node_type}: {node_id}'
                       f'\n   parent: {parent}'
                       f'\n   prev: {prev}'
                       f'\n   next: {next}')
                if group:
                    msg += (f'\n   head: {head}'
                            f'\n   tail: {tail}')
                print(msg)

        rdf.OscFunc(
            lambda msg, *_: action(*msg),
            '/n_info', self.server.addr,
            arg_template=[self.node_id]).one_shot()
        self.server.send_msg('/n_query', self.node_id)

    def register(self, playing=True, running=True):
        self.server._node_watcher.register(self, playing, running)

    def unregister(self):
        self.server._node_watcher.unregister(self)

    def on_free(self, action):
        def action_wrapper():
            fn.value(action, self)
            mdl.NotificationCenter.unregister(self, '/n_end', self)

        self.register()
        mdl.NotificationCenter.register(self, '/n_end', self, action_wrapper)

    def wait_for_free(self):
        condition = stm.Condition()
        self.on_free(lambda: condition.unhang())
        condition.hang()

    def move_before(self, node):
        self.group = node.group
        self.server.send_msg('/n_before', self.node_id, node.node_id) # 18

    def move_before_msg(self, node):
        self.group = node.group # TODO: estos msg podrían tener un parámetros update=True por defecto, pero no sé dónde se usan estas funciones aún.
        return ['/n_before', self.node_id, node.node_id] # 18

    def move_after(self, node):
        self.group = node.group
        self.server.send_msg('/n_after', self.node_id, node.node_id) # 19

    def move_after_msg(self, node):
        self.group = node.group
        return ['/n_after', self.node_id, node.node_id] # 19

    def move_to_head(self, group=None):
        group = group or self.server.default_group
        group.move_node_to_head(self) # se implementa en AbstractGroup

    def move_to_head_msg(self, group=None):
        group = group or self.server.default_group
        return group.move_node_to_head_msg(self) # se implementa en AbstractGroup

    def move_to_tail(self, group=None):
        group = group or self.server.default_group
        group.move_node_to_tail(self) # se implementa en AbstractGroup

    def move_to_tail_msg(self, group=None):
        group = group or self.server.default_group
        return group.move_node_to_tail_msg(self) # se implementa en AbstractGroup

    def order_nodes_msg(self, nodes):
        msg = ['/n_after'] # 19
        for first, to_move_after in utl.pairwise(nodes):
            msg.append(to_move_after.node_id)
            msg.append(first.node_id)
        return msg

    # TODO: VER:
    # ==
    # hash
    # printOn


    ### Node parameter interface ###

    def _as_control_input(self):
        return self.node_id

    def _as_target(self):
        return self


# // common base for Group and ParGroup classes
class AbstractGroup(Node):
    def __init__(self, target=None, add_action='addToHead', register=False):
        # // Immediately sends.
        super().__init__()
        target = gpp.node_param(target)._as_target()
        self.server = target.server
        self.node_id = self.server.next_node_id()
        add_action_id = type(self).add_actions[add_action]
        if add_action_id < 2:
            self.group = target
        else:
            self.group = target.group
        self._init_register(register)
        self.server.send_msg(
            self.creation_cmd(), self.node_id,
            add_action_id, target.node_id)

    def new_msg(self, target=None, add_action='addToHead'):
        # // if target is nil set to default group of server specified when basicNew was called
        target = gpp.node_param(target)._as_target()
        add_action_id = type(self).add_actions[add_action]
        if add_action_id < 2:
            self.group = target
        else:
            self.group = target.group
        return [self.creation_cmd(), self.node_id,
                add_action_id, target.node_id]

    # // for bundling
    def add_to_head_msg(self, group=None):
        # // if group is nil set to default group of server specified when basicNew was called
        self.group = group or self.server.default_group
        return [self.creation_cmd(), self.node_id, 0, self.group.node_id]

    def add_to_tail_msg(self, group=None):
        # // if group is nil set to default group of server specified when basicNew was called
        self.group = group or self.server.default_group
        return [self.creation_cmd(), self.node_id, 1, self.group.node_id]

    def add_after_msg(self, node):
        self.group = node.group
        return [self.creation_cmd(), self.node_id, 3, self.group.node_id]

    def add_before_msg(self, node):
        self.group = node.group
        return [self.creation_cmd(), self.node_id, 2, self.group.node_id]

    def add_replace_msg(self, node_to_replace):
        self.group = node_to_replace.group
        return [self.creation_cmd(), self.node_id, 4, node_to_replace.node_id]

    @classmethod
    def after(cls, node):
        return cls(node, 'addAfter')

    @classmethod
    def before(cls, node):
        return cls(node, 'addBefore')

    @classmethod
    def head(cls, group):
        return cls(group, 'addToHead')

    @classmethod
    def tail(cls, group):
        return cls(group, 'addToTail')

    @classmethod
    def replace(cls, node_to_replace):
        return cls(node_to_replace, 'addReplace')

    # // move Nodes to this group

    def move_node_to_head(self, node):
        node.group = self
        self.server.send_msg('/g_head', self.node_id, node.node_id) # 22

    def move_node_to_head_msg(self, node):
        node.group = self
        return ['/g_head', self.node_id, node.node_id] # 22

    def move_node_to_tail(self, node):
        node.group = self
        self.server.send_msg('/g_tail', self.node_id, node.node_id) # 23

    def move_node_to_tail_msg(self, node):
        node.group = self
        return ['/g_tail', self.node_id, node.node_id] # 23

    def free_all(self):
        # // free my children, but this node is still playing
        self.server.send_msg('/g_freeAll', self.node_id) # 24

    def free_all_msg(self):
        # // free my children, but this node is still playing
        return ['/g_freeAll', self.node_id] # 24

    def deep_free(self):
        self.server.send_msg('/g_deepFree', self.node_id) # 50

    def deep_free_msg(self):
        return ['/g_deepFree', self.node_id] # 50

    def dump_tree(self, controls=False):
        '''
        Ask the server to dump this node tree to stdout.

        Parameters
        ----------
        controls: bool
            If `True` also print synth controls with current values.
        '''

        self.server.send_msg('/g_dumpTree', self.node_id, int(controls))

    def query_tree(self, controls=False, action=None, timeout=3):
        '''
        Query the groups's node tree.

        Parameters
        ----------
        controls: bool
            If `True` also request synth controls values.
        action: function
            A responder function that receives the data in JSON format.
        timeout: int | float
            Request timeout in seconds.
        '''

        done = False

        def resp_func(msg, *_):
            print_controls = bool(msg[1])
            key_name = f'Group({msg[2]})'
            outdct = dict()
            outdct[key_name] = dict()
            i = 2

            if msg[3] > 0:
                def dump_func(outdct, num_children):
                    nonlocal i
                    for _ in range(num_children):
                        if msg[i + 1] >= 0:
                            i += 2
                        else:
                            if print_controls:
                                i += msg[i + 3] * 2 + 1
                            i += 3
                        node_id = f'{msg[i]}'
                        if msg[i + 1] >= 0:
                            key_name = f'Group({node_id})'
                            outdct[key_name] = dict()
                            if msg[i + 1] > 0:
                                dump_func(outdct[key_name], msg[i + 1])
                        else:
                            key_name = f'Synth({node_id}, {msg[i + 2]})'
                            outdct[key_name] = dict()
                            if print_controls:
                                j = 0
                                for _ in range(msg[i + 3]):
                                    outdct[key_name][f'{msg[i + 4 + j]}'] =\
                                        msg[i + 5 + j]
                                    j += 2

                dump_func(outdct[key_name], msg[3])

            nonlocal done
            done = True

            if action:
                fn.value(action, outdct)
            else:
                self._pretty_tree(outdct)

        resp = rdf.OscFunc(resp_func, '/g_queryTree.reply', self.server.addr)
        resp.one_shot()

        def timeout_func():
            if not done:
                resp.free()
                _logger.warning(
                    f"server '{self.server.name}' failed to respond "
                    f"to '/g_queryTree' after {timeout} seconds")

        self.server.send_msg('/g_queryTree', self.node_id, int(controls))
        clk.SystemClock.sched(timeout, timeout_func)

    def _pretty_tree(self, d, indent=0, tab=2):
        for key, value in d.items():
            _logger.info(' ' * tab * indent + str(key))
            if key.startswith('Synth'):
                log = ' ' * tab * (indent + 1)
                for k, v in value.items():
                    log += f'{k}: {v} '
                _logger.info(log)
            elif isinstance(value, dict):
                self._pretty_tree(value, indent+1)

    @staticmethod
    def creation_cmd():
        raise NotImplementedError()

    def __repr__(self):
        return f'{type(self).__name__}({self.node_id})'


class Group(AbstractGroup):
    """Client-side representation of a group node on the server."""

    @staticmethod
    def creation_cmd():
        return '/g_new' # 21


class ParGroup(AbstractGroup):
    @staticmethod
    def creation_cmd():
        return '/p_new' # 63


class RootNode(Group):
    """Persistent root group on the server."""

    roots = dict()

    def __new__(cls, server=None):
        server = server or srv.Server.default
        if server.name in cls.roots:
            return cls.roots[server.name]
        else:
            obj = super(gpp.NodeParameter, cls).__new__(cls)
            obj.server = server
            obj.node_id = 0
            obj._is_playing = True  # Always true even if not watched.
            obj._is_running = True  # Always true even if not watched.
            obj.group = obj
            cls.roots[obj.server.name] = obj
            return obj

    def __init__(self, _=None):
        super(gpp.NodeParameter, self).__init__(self)

    def run(self):
        _logger.warning('run has no effect on RootNode')

    def free(self):
        _logger.warning('free has no effect on RootNode')

    def move_before(self):
        _logger.warning('moveBefore has no effect on RootNode')

    def move_after(self):
        _logger.warning('move_after has no effect on RootNode')

    def move_to_head(self):
        _logger.warning('move_to_head has no effect on RootNode')

    def move_to_tail(self):
        _logger.warning('move_to_tail has no effect on RootNode')

    @classmethod
    def free_all_roots(cls):  # renamed because collision with instance method
        for rn in cls.roots.values():
            rn.free_all()


class Synth(Node):
    """Client-side representation of a synth node on the server."""

    def __init__(self, def_name, args=None, target=None,
                 add_action='addToHead', register=False):
        # // Immediately sends.
        super().__init__()
        target = gpp.node_param(target)._as_target()
        self.server = target.server
        self.node_id = self.server.next_node_id()
        add_action_id = type(self).add_actions[add_action]
        if add_action_id < 2:
            self.group = target
        else:
            self.group = target.group
        self.def_name = def_name
        self._init_register(register)
        self.server.send_msg(
            '/s_new', # 9
            self.def_name, self.node_id,
            add_action_id, target.node_id,
            *gpp.node_param(args or [])._as_osc_arg_list())

    # // does not send (used for bundling)
    @classmethod
    def basic_new(cls, def_name, server=None, node_id=None):
        obj = super().basic_new(server, node_id)
        obj.def_name = def_name
        return obj

    @classmethod
    def new_paused(cls, def_name, args=None, target=None, add_action='addToHead'):
        target = gpp.node_param(target)._as_target()
        server = target.server
        add_action_id = cls.add_actions[add_action]
        synth = cls.basic_new(def_name, server)
        if add_action_id < 2:
            synth.group = target
        else:
            synth.group = target.group
        synth.server.send_bundle(
            0,
            [
                '/s_new', # 9
                synth.def_name, synth.node_id,
                add_action_id, target.node_id,
                *gpp.node_param(args or [])._as_osc_arg_list()
            ],
            [
                '/n_run', # 12
                synth.node_id, 0
            ]
        )
        return synth

    @classmethod
    def new_replace(cls, node_to_replace, def_name, args=None, same_id=False): # BUG: renombrado porque no pueden haber métodos de instancia y clase con el mismo nombre.
        if same_id:
            new_node_id = node_to_replace.node_id
        else:
            new_node_id = None
        server = node_to_replace.server
        synth = cls.basic_new(def_name, server, new_node_id)
        synth.server.send_msg(
            '/s_new', # 9
            synth.def_name, synth.node_id,
            4, node_to_replace.node_id, # 4 -> 'addReplace'
            *gpp.node_param(args or [])._as_osc_arg_list())
        return synth

    # node_id -1
    @classmethod # TODO: este tal vez debería ir arriba
    def grain(cls, def_name, args=None, target=None, add_action='addToHead'):
        target = gpp.node_param(target)._as_target()
        server = target.server
        server.send_msg(
            '/s_new', # 9
            def_name.as_def_name(), -1, # BUG: as_def_name no está implementado puede ser método de Object
            cls.add_actions[add_action], target.node_id,
            *gpp.node_param(args or [])._as_osc_arg_list())

    def new_msg(self, target=None, args=None, add_action='addToHead'):
        add_action_id = self.add_actions[add_action]
        target = gpp.node_param(target)._as_target()
        if add_action_id < 2:
            self.group = target
        else:
            self.group = target.group
        return [
            '/s_new', self.def_name, self.node_id, add_action_id,
            target.node_id, *gpp.node_param(args or [])._as_osc_arg_list()] # 9

    @classmethod
    def after(cls, node, def_name, args=None):
        return cls(def_name, args, node, 'addAfter')

    @classmethod
    def before(cls, node, def_name, args=None):
        return cls(def_name, args, node, 'addBefore')

    @classmethod
    def head(cls, group, def_name, args=None):
        return cls(def_name, args, group, 'addToHead')

    @classmethod
    def tail(cls, group, def_name, args=None):
        return cls(def_name, args, group, 'addToTail')

    def replace(self, def_name, args=None, same_id=False):
        return type(self).new_replace(self, def_name, args, same_id)

    # // for bundling
    def add_to_head_msg(self, group, args):
        # // If group is nil set to default group of the
        # // server specified when basicNew was called.
        if group is not None:
            self.group = group
        else:
            self.group = self.server.default_group
        return [
            '/s_new', self.def_name, self.node_id, 0, self.group.node_id,
            *gpp.node_param(args)._as_osc_arg_list()] # 9

    def add_to_tail_msg(self, group, args):
        # // if aGroup is nil set to default group of server specified when basicNew was called
        if group is not None:
            self.group = group
        else:
            self.group = self.server.default_group
        return [
            '/s_new', self.def_name, self.node_id, 1, self.group.node_id,
            *gpp.node_param(args)._as_osc_arg_list()] # 9

    def add_after_msg(self, node, args=None):
        self.group = node.group
        return [
            '/s_new', self.def_name, self.node_id, 3, node.node_id,
            *gpp.node_param(args or [])._as_osc_arg_list()] # 9

    def add_before_msg(self, node, args=None):
        self.group = node.group
        return [
            '/s_new', self.def_name, self.node_id, 2, node.node_id,
            *gpp.node_param(args or [])._as_osc_arg_list()] # 9

    def add_replace_msg(self, node_to_replace, args):
        self.group = node_to_replace.group
        return [
            '/s_new', self.def_name, self.node_id, 4, node_to_replace.node_id,
            *gpp.node_param(args)._as_osc_arg_list()] # 9

    def get(self, index, action):
        def resp_func(msg, *_):
            # // The server replies with a message of the
            # // form: [/n_set, node ID, index, value].
            # // We want 'value' which is at index 3.
            fn.value(action, msg[3])

        rdf.OscFunc(
            resp_func, '/n_set', self.server.addr,
            arg_template=[self.node_id, index]).one_shot()

        self.server.send_msg('/s_get', self.node_id, index)  # 44

    def get_msg(self, index):
        return ['/s_get', self.node_id, index] # 44

    def getn(self, index, count, action):
        def resp_func(msg, *_):
            # // The server replies with a message of the form
            # // [/n_setn, node ID, index, count, *values].
            # // We want '*values' which are at indexes 4 and above.
            fn.value(action, msg[4:])

        rdf.OscFunc(
            resp_func, '/n_setn', self.server.addr,
            arg_template=[self.node_id, index]).one_shot()

        self.server.send_msg('/s_getn', self.node_id, index)  # 45

    def getn_msg(self, index, count):
        return ['/s_getn', self.node_id, index, count] # 45

    def seti(self, *args): # // args are [key, index, value, key, index, value ...]
        osc_msg = []
        synth_desc = sdc.SynthDescLib.at(self.def_name)
        if synth_desc is None:
            _logger.warning(
                f"message seti failed, SynthDef '{self.def_name}' "
                "not found in SynthDescLib")
            return
        for key, offset, value in utl.gen_cclumps(args, 3):
            if key in synth_desc.control_dict:
                cname = synth_desc.control_dict[key]
                if offset < cname.num_channels:
                    osc_msg.append(cname.index + offset)
                    if isinstance(value, list):
                        osc_msg.append(value[:cname.num_channels - offset]) # keep
                    else:
                        osc_msg.append(value)
        self.server.send_msg(
            '/n_set', self.node_id,
            *gpp.node_param(osc_msg)._as_osc_arg_list())

    def __repr__(self):
        return f'{type(self).__name__}({self.def_name} : {self.node_id})'



"""EnvGen.sc"""

from ...base import utils as utl
from .. import ugen as ugn
from .. import envelope as evp


class Done(ugn.UGen):
    _default_rate = 'control'

    NONE = 0
    PAUSE_SELF = 1
    FREE_SELF = 2
    FREE_SELF_AND_PREV = 3
    FREE_SELF_AND_NEXT = 4
    FREE_SELF_AND_FREE_ALL_IN_PREV = 5
    FREE_SELF_AND_FREE_ALL_IN_NEXT = 6
    FREE_SELF_TO_HEAD = 7
    FREE_SELF_TO_TAIL = 8
    FREE_SELF_PAUSE_PREV = 9
    FREE_SELF_PAUSE_NEXT = 10
    FREE_SELF_AND_DEEP_FREE_PREV = 11
    FREE_SELF_AND_DEEP_FREE_NEXT = 12
    FREE_ALL_IN_GROUP = 13
    FREE_GROUP = 14
    FREE_SELF_RESUME_NEXT = 15

    @classmethod
    def kr(cls, src):
        return cls._multi_new('control', src)


class NodeControlUGen(ugn.UGen):
    _default_rate = 'control'


class FreeSelf(NodeControlUGen):
    @classmethod
    def kr(cls, input):
        cls._multi_new('control', input)
        return input


class PauseSelf(NodeControlUGen):
    @classmethod
    def kr(cls, input):
        cls._multi_new('control', input)
        return input


class FreeSelfWhenDone(NodeControlUGen):
    @classmethod
    def kr(cls, src):
        return cls._multi_new('control', src)


class PauseSelfWhenDone(NodeControlUGen):
    @classmethod
    def kr(cls, src):
        return cls._multi_new('control', src)


class Pause(NodeControlUGen):
    @classmethod
    def kr(cls, gate, id):
        return cls._multi_new('control', gate, id)


class Free(NodeControlUGen):
    @classmethod
    def kr(cls, trig, id):
        return cls._multi_new('control', trig, id)


class EnvGen(ugn.UGen):
    @classmethod
    def ar(cls, env, gate=1.0, level_scale=1.0, level_bias=0.0,
           time_scale=1.0, done_action=0):
        '''
        ``env`` can be a tuple, a list of tuples for multiple channels
        or an instance of Env.
        '''
        if isinstance(env, evp.Env):
            env = utl.unbubble(env.envgen_format())  # Was asMultichannelArray.
        return cls._multi_new('audio', gate, level_scale, level_bias,
                              time_scale, done_action, env)

    @classmethod
    def kr(cls, env, gate=1.0, level_scale=1.0, level_bias=0.0,
           time_scale=1.0, done_action=0):
        '''
        env can be a tuple, a list of tuples for multiple channels
        or an instance of Env.
        '''
        if isinstance(env, evp.Env):
            env = utl.unbubble(env.envgen_format())  # Was asMultichannelArray.
        return cls._multi_new('control', gate, level_scale, level_bias,
                              time_scale, done_action, env)

    @classmethod
    def _new1(cls, rate, *args):
        obj = cls._create_ugen_object(rate)
        obj._add_to_synth()
        args = list(args)
        env = args.pop()
        return obj._init_ugen(*args, *env)

    # Override may be an optimization in sclang.
    # def _init_ugen(self, inputs)  # override

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.


class Linen(ugn.UGen):
    _default_rate = 'control'

    @classmethod
    def kr(cls, gate=1.0, attack_time=0.01, sus_level=1.0,
           release_time=1.0, done_action=0):
        return cls._multi_new('control', gate, attack_time, sus_level,
                              release_time, done_action)



"""
Special Index and Symbols

Indexes from enum in Opcodes.h, maps to Symbol in PyrParseNode.cpp
initSpecialSelectors(). Names  of index opcodes (Opcodes.h) are selectors names
(PyrParseNode). Used by BasicOpUGens to get an ID number for the operator.
"""

# TODO: special_index se usa en BasicOpUGen pero, además, creo, el atributo
# operator tiene que ser uno de estos símbolos (porque un y bin comparten
# números). Estos atributos special_index y name se usan en write_def de UGen.

# TODO: Verify precise behavior and synonyms

# unary opcodes list
_unops_list = [
    ('neg', '__neg__'),  # '-' unary
    ('not',),  # boolean not buleano in sclang
    ('isNil',),  # nil/obj in sclang
    ('notNil',),  # nil/obj in sclang
    ('bitNot', '__invert__', 'invert'),  # ~
    ('abs', '__abs__'),  # abs()
    ('asFloat',),
    ('asInt',),
    ('ceil', '__ceil__'),  # underscores?
    ('floor', '__floor__'),  # underscores?
    ('frac',),  # is math.modf(x)[0], not same behavior, returns tuple
    ('sign',),
    ('squared',),
    ('cubed',),
    ('sqrt',),  # math.sqrt(x), radix e
    ('exp',),  # e to the power of the receiver, math.exp(x)
    ('reciprocal',),
    ('midicps',),
    ('cpsmidi',),
    ('midiratio',),
    ('ratiomidi',),
    ('dbamp',),
    ('ampdb',),
    ('octcps',),
    ('cpsoct',),
    ('log',),  # math.log
    ('log2',),
    ('log10',),
    ('sin',),  # math.sin
    ('cos',),  # etc...
    ('tan',),
    ('asin',),
    ('acos',),
    ('atan',),
    ('sinh',),
    ('cosh',),
    ('tanh',),
    ('rand',),
    ('rand2',),
    ('linrand',),
    ('bilinrand',),
    #'exprand', # existe en AbstractFunction, está comentado acá.
    #'biexprand',
    ('sum3rand',),
    #'gammarand',
    #'gaussrand',
    #'poirand',
    ('distort',),
    ('softclip',),
    ('coin',),
    ('digitValue',), # no existe en sclang
    ('silence',), # no existe en sclang
    ('thru',), # no existe en sclang
    ('rectWindow', 'rectwindow'),
    ('hanWindow', 'hanwindow'),
    ('welWindow', 'welwindow'),
    ('triWindow', 'triwindow'),
    ('ramp',),
    ('scurve',),
    #opNumUnarySelectors
]

# binary opcodes list
_binops_list = [
    ('+', '__add__', '__radd__', 'add'),
    ('-', '__sub__', '__rsub__', 'sub'),
    ('*', '__mul__', '__rmul__', 'mul'),
    ('div', '__floordiv__', '__floordiv__', 'floordiv'),  # Python's '//'
    ('/', '__truediv__', '__rtruediv__', 'truediv'),
    ('mod', '__mod__', '__rmod__'),
    ('==', '__eq__', 'eq'),
    ('!=', '__ne__', 'ne'),
    ('<', '__lt__', 'lt'),
    ('>', '__gt__', 'gt'),
    ('<=', '__le__', 'le'),
    ('>=', '__ge__', 'ge'),
    #'===', # Python is
    #'!==', # Python is not
    ('min',),
    ('max',),
    ('bitAnd', 'bitand', '__and__', '__rand__', 'and'),
    ('bitOr', 'bitor', '__or__', '__ror__', 'or'),
    ('bitXor', 'bitxor', '__xor__', '__rxor__', 'xor'),
    ('lcm',),
    ('gcd',),
    ('round',), # Es trunc(x, quant) En Python __round__ es Unario con argumento
    ('roundUp', 'roundup'),
    ('trunc',), # BUG: Es trunc(x, quant) # En Python __truc__ es Unario, es la operación por defecto para int(x)
    ('atan2',),
    ('hypot',),
    ('hypotApx', 'hypotx'),
    ('pow', '__pow__', '__rpow__'),  # **
    ('leftShift', 'leftshift', '__lshift__', '__rlshift__', 'lshift'),
    ('rightShift', 'rightshift', '__rshift__', '__rrshift__', 'rshift'),
    ('unsignedRightShift', 'urightshift'),
    ('fill',),
    ('ring1',), # a * (b + 1) == a * b + a
    ('ring2',), # a * b + a + b
    ('ring3',), # a*a*b
    ('ring4',), # a*a*b - a*b*b
    ('difsqr',), # a*a - b*b
    ('sumsqr',), # a*a + b*b
    ('sqrsum',), # (a + b)^2
    ('sqrdif',), # (a - b)^2
    ('absdif',), # |a - b|
    ('thresh',),
    ('amclip',),
    ('scaleneg',),
    ('clip2',),
    ('excess',),
    ('fold2',),
    ('wrap2',),
    ('firstArg', 'first_arg'),
    ('rrand',),
    ('exprand',),
    #opNumBinarySelectors
]


def _build_op_dict(oplist):
    ret = dict()
    for i, item in enumerate(oplist):
        for name in item:
            ret[name] = [i, item[0]] # key: [special_index, sc_name]
    return ret


_unops = _build_op_dict(_unops_list)
_binops = _build_op_dict(_binops_list)


# Interface

def sc_spindex_opname(operator):
    '''
    Returns (special_index, sc_opname) or (-1, None). operator is a str.
    '''
    try: return _unops[operator]
    except KeyError: pass
    try: return _binops[operator]
    except KeyError: pass
    return (-1, None)


def special_index(operator):
    '''Returns server operators special index or -1.'''
    ret = sc_spindex_opname(operator)
    return ret[0]


def sc_opname(operator):
    '''Returns server operator name or None.'''
    ret = sc_spindex_opname(operator)
    return ret[1]


def sc_opname_from_index(index, arity='unary'):
    if arity == 'unary':
        return _unops_list[index][0]
    elif arity == 'binary':
        return _binops_list[index][0]
    else:
        raise ValueError(f'arity {arity} not supported by special index')



"""FFTUnpacking.sc"""

# // "Unpack FFT" UGens (c) 2007 Dan Stowell.
# // Magical UGens for treating FFT data as demand-rate streams.

from .. import ugen as ugn


# // Actually this just wraps up a bundle of Unpack1FFT UGens.
class UnpackFFT(ugn.MultiOutUGen):
    ...


class Unpack1FFT(ugn.UGen):
    ...


# // Conveniences to apply calculations to an FFT chain.
class PV_ChainUGen(ugn.WidthFirstUGen):
    ...


# // This does the demanding, to push the data back into an FFT buffer.
class PackFFT(PV_ChainUGen):
    ...




import argparse
import logging
import sys
import runpy

import sc3


parser = argparse.ArgumentParser(prog='sc3')

parser.add_argument(
    '-v', '--version', action='version',
    version=f'%(prog)s {sc3.__version__}')
parser.add_argument(
    '-N', '--nrt', action='store_const', const='nrt',
    help='non real time mode')
parser.add_argument(
    '-u', '--udp-port', type=int, default=57120,
    help='udp library port')
parser.add_argument(
    '-r', '--port-range', type=int, default=10,
    help='udp available port range')
parser.add_argument(
    '-s', '--setup-file', type=str, default=None,
    help='library setup.py file')
parser.add_argument(
    '-V', '--verbosity', type=int, default=20,
    help='logger level')
parser.add_argument('file', nargs='?')
parser.add_argument('args', nargs=argparse.REMAINDER)

args = parser.parse_args()


### Config library ###

sc3.LIB_MODE = args.nrt or 'rt'
sc3.LIB_PORT = args.udp_port
sc3.LIB_PORT_RANGE = args.port_range
sc3.LIB_SETUP_FILE = args.setup_file


### Init library ###

sc3.init(sc3.LIB_MODE, logging.getLevelName(args.verbosity))


### Run script ###

if args.file:
    sys.argv = [args.file, *args.args]  # Simulate script args.
    runpy.run_path(args.file)



"""
The content of this file is a modified version of python-osc from

https://github.com/attwad/python-osc

whose License is as follows:

This is free and unencumbered software released into the public domain.

Anyone is free to copy, modify, publish, use, compile, sell, or
distribute this software, either in source code form or as a compiled
binary, for any purpose, commercial or non-commercial, and by any
means.

In jurisdictions that recognize copyright laws, the author or authors
of this software dedicate any and all copyright interest in the
software to the public domain. We make this dedication for the benefit
of the public at large and to the detriment of our heirs and
successors. We intend this dedication to be an overt act of
relinquishment in perpetuity of all present and future rights to this
software under copyright law.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

For more information, please refer to <http://unlicense.org/>
"""


import struct
import socketserver
import sys
import logging
import collections
from typing import Union, Tuple, Any, Iterator, List

from . import main as _libsc3


_logger = logging.getLogger(__name__)


### OSC Exceptions ###


class OscParseError(Exception):
    pass


class OscTypeParseError(OscParseError):
    pass


class OscMessageParseError(OscParseError):
    pass


class OscBundleParseError(OscParseError):
    pass


class OscBuildError(Exception):
    pass


class OscTypeBuildError(OscBuildError):
    pass


class OscMessageBuildError(OscBuildError):
    pass


class OscBundleBuildError(OscBuildError):
    pass


### Constants ###


IMMEDIATELY = 1

_IMMEDIATELY_DGRAM = struct.pack('>Q', 1)
_BUNDLE_PREFIX_DGRAM = b'#bundle\x00'
_EMPTY_STR_DGRAM = b'\x00\x00\x00\x00'

_INT_DGRAM_LEN = 4
_FLOAT_DGRAM_LEN = 4
_DOUBLE_DGRAM_LEN = 8
_TIMETAG_DGRAM_LEN = 8

# Strings and blob dgram length is always a multiple of 4 bytes.
_STRING_DGRAM_PAD = 4
_BLOB_DGRAM_PAD = 4


### OSC Types ###


def write_string(val: str) -> bytes:
    """Returns the OSC string equivalent of the given python string.

    Raises:
      - BuildError if the string could not be encoded.
    """
    try:
        dgram = val.encode('utf-8')  # Default, but better be explicit.
    except (UnicodeEncodeError, AttributeError) as e:
        raise OscTypeBuildError('Incorrect string, could not encode') from e
    diff = _STRING_DGRAM_PAD - (len(dgram) % _STRING_DGRAM_PAD)
    dgram += (b'\x00' * diff)
    return dgram


def get_string(dgram: bytes, start_index: int) -> Tuple[str, int]:
    """Get a python string from the datagram, starting at pos start_index.

    According to the specifications, a string is:
    "A sequence of non-null ASCII characters followed by a null,
    followed by 0-3 additional null characters to make the total number
    of bits a multiple of 32".

    Args:
      dgram: A datagram packet.
      start_index: An index where the string starts in the datagram.

    Returns:
      A tuple containing the string and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    if start_index < 0:
        raise OscTypeParseError('start_index < 0')
    offset = 0
    try:
        if (len(dgram) > start_index + _STRING_DGRAM_PAD
                and dgram[start_index + _STRING_DGRAM_PAD] == _EMPTY_STR_DGRAM):
            return '', start_index + _STRING_DGRAM_PAD
        while dgram[start_index + offset] != 0:
            offset += 1
        # Align to a byte word.
        if (offset) % _STRING_DGRAM_PAD == 0:
            offset += _STRING_DGRAM_PAD
        else:
            offset += (-offset % _STRING_DGRAM_PAD)
        # Python slices do not raise an IndexError past the last index,
        # do it ourselves.
        if offset > len(dgram[start_index:]):
            raise OscTypeParseError('Datagram is too short')
        data_str = dgram[start_index:start_index + offset]
        return data_str.replace(b'\x00', b'').decode('utf-8'), start_index + offset
    except IndexError as e:
        raise OscTypeParseError('Could not parse datagram') from e
    except TypeError as e:
        raise OscTypeParseError('Could not parse datagram') from e


def write_int(val: int) -> bytes:
    """Returns the datagram for the given integer parameter value

    Raises:
      - BuildError if the int could not be converted.
    """
    try:
        return struct.pack('>i', val)
    except struct.error as e:
        raise OscTypeBuildError('Wrong argument value passed') from e


def get_int(dgram: bytes, start_index: int) -> Tuple[int, int]:
    """Get a 32-bit big-endian two's complement integer from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the integer starts in the datagram.

    Returns:
      A tuple containing the integer and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _INT_DGRAM_LEN:
            raise OscTypeParseError('Datagram is too short')
        return (
            struct.unpack(
                '>i', dgram[start_index:start_index + _INT_DGRAM_LEN])[0],
            start_index + _INT_DGRAM_LEN)
    except (struct.error, TypeError) as e:
        raise OscTypeParseError('Could not parse datagram') from e


def write_timetag(timetag: int) -> bytes:
    try:
        return struct.pack('>Q', timetag)
    except struct.error as e:
        raise OscTypeBuildError('Wrong argument value passed') from e


def get_timetag(dgram: bytes, start_index: int) -> Tuple[int, int]:
    """Get a 64-bit big-endian unsigned integer from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the integer starts in the datagram.

    Returns:
      Time as unsigned int OSC timetag.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _TIMETAG_DGRAM_LEN:
            raise OscTypeParseError('Datagram is too short')
        return (
            struct.unpack(
                '>Q', dgram[start_index:start_index + _TIMETAG_DGRAM_LEN])[0],
            start_index + _TIMETAG_DGRAM_LEN)
    except (struct.error, TypeError) as e:
        raise OscTypeParseError('Could not parse datagram') from e


def write_float(val: float) -> bytes:
    """Returns the datagram for the given float parameter value

    Raises:
      - BuildError if the float could not be converted.
    """
    try:
        return struct.pack('>f', val)
    except struct.error as e:
        raise OscTypeBuildError('Wrong argument value passed') from e


def get_float(dgram: bytes, start_index: int) -> Tuple[float, int]:
    """Get a 32-bit big-endian IEEE 754 floating point number from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the float starts in the datagram.

    Returns:
      A tuple containing the float and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _FLOAT_DGRAM_LEN:
            # Noticed that Reaktor doesn't send the last bunch of \x00 needed to make
            # the float representation complete in some cases, thus we pad here to
            # account for that.
            dgram = dgram + b'\x00' * (_FLOAT_DGRAM_LEN - len(dgram[start_index:]))
        return (
            struct.unpack(
                '>f', dgram[start_index:start_index + _FLOAT_DGRAM_LEN])[0],
            start_index + _FLOAT_DGRAM_LEN)
    except (struct.error, TypeError) as e:
        raise OscTypeParseError('Could not parse datagram') from e


def write_double(val: float) -> bytes:
    """Returns the datagram for the given double parameter value

    Raises:
      - BuildError if the double could not be converted.
    """
    try:
        return struct.pack('>d', val)
    except struct.error as e:
        raise OscTypeBuildError('Wrong argument value passed') from e


def get_double(dgram: bytes, start_index: int) -> Tuple[float, int]:
    """Get a 64-bit big-endian IEEE 754 floating point number from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the double starts in the datagram.

    Returns:
      A tuple containing the double and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _DOUBLE_DGRAM_LEN:
            raise ParseError('Datagram is too short')
        return (
            struct.unpack(
                '>d', dgram[start_index:start_index + _DOUBLE_DGRAM_LEN])[0],
            start_index + _DOUBLE_DGRAM_LEN)
    except (struct.error, TypeError) as e:
        raise OscTypeParseError('Could not parse datagram') from e


def write_blob(val: bytes) -> bytes:
    """Returns the datagram for the given blob parameter value.

    Raises:
      - BuildError if the value was empty or if its size didn't fit an OSC int.
    """
    if not val:
        raise OscTypeBuildError('Blob value cannot be empty')
    dgram = write_int(len(val))
    dgram += val
    while len(dgram) % _BLOB_DGRAM_PAD != 0:
        dgram += b'\x00'
    return dgram


def get_blob(dgram: bytes, start_index: int) -> Tuple[bytes, int]:
    """ Get a blob from the datagram.

    According to the specifications, a blob is made of
    "an int32 size count, followed by that many 8-bit bytes of arbitrary
    binary data, followed by 0-3 additional zero bytes to make the total
    number of bits a multiple of 32".

    Args:
      dgram: A datagram packet.
      start_index: An index where the float starts in the datagram.

    Returns:
      A tuple containing the blob and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    size, int_offset = get_int(dgram, start_index)
    # Make the size a multiple of 32 bits.
    total_size = size + (-size % _BLOB_DGRAM_PAD)
    end_index = int_offset + size
    if end_index - start_index > len(dgram[start_index:]):
        raise OscTypeParseError('Datagram is too short')
    return dgram[int_offset:int_offset + size], int_offset + total_size


def write_rgba(val: bytes) -> bytes:
    """Returns the datagram for the given rgba32 parameter value

    Raises:
      - BuildError if the int could not be converted.
    """
    try:
        return struct.pack('>I', val)
    except struct.error as e:
        raise OscTypeBuildError('Wrong argument value passed') from e


def get_rgba(dgram: bytes, start_index: int) -> Tuple[bytes, int]:
    """Get an rgba32 integer from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the integer starts in the datagram.

    Returns:
      A tuple containing the integer and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _INT_DGRAM_LEN:
            raise OscTypeParseError('Datagram is too short')
        return (
            struct.unpack(
                '>I', dgram[start_index:start_index + _INT_DGRAM_LEN])[0],
            start_index + _INT_DGRAM_LEN)
    except (struct.error, TypeError) as e:
        raise OscTypeParseError('Could not parse datagram') from e


def write_midi(val: Tuple[Tuple[int, int, int, int], int]) -> bytes:
    """Returns the datagram for the given MIDI message parameter value

       A valid MIDI message: (port id, status byte, data1, data2).

    Raises:
      - BuildError if the MIDI message could not be converted.
    """
    if len(val) != 4:
        raise OscTypeBuildError('MIDI message length is invalid')
    try:
        value = sum(
            (value & 0xFF) << 8 * (3 - pos) for pos, value in enumerate(val))
        return struct.pack('>I', value)
    except struct.error as e:
        raise OscTypeBuildError('Wrong argument value passed') from e


def get_midi(dgram: bytes, start_index: int) -> Tuple[Tuple[int, int, int, int], int]:
    """Get a MIDI message (port id, status byte, data1, data2) from the datagram.

    Args:
      dgram: A datagram packet.
      start_index: An index where the MIDI message starts in the datagram.

    Returns:
      A tuple containing the MIDI message and the new end index.

    Raises:
      ParseError if the datagram could not be parsed.
    """
    try:
        if len(dgram[start_index:]) < _INT_DGRAM_LEN:
            raise OscTypeParseError('Datagram is too short')
        val = struct.unpack(
            '>I', dgram[start_index:start_index + _INT_DGRAM_LEN])[0]
        midi_msg = tuple(
            (val & 0xFF << 8 * i) >> 8 * i for i in range(3, -1, -1))
        return (midi_msg, start_index + _INT_DGRAM_LEN)
    except (struct.error, TypeError) as e:
        raise OscTypeParseError('Could not parse datagram') from e


### OSC Bundle ###


class OscBundle(object):
    """Bundles elements that should be triggered at the same time.

    An element can be another OscBundle or an OscMessage.
    """

    def __init__(self, dgram: bytes):
        """Initializes the OscBundle with the given datagram.

        Args:
          dgram: a UDP datagram representing an OscBundle.
        Raises:
          ParseError: if the datagram could not be parsed into an OscBundle.
        """
        # Interesting stuff starts after the initial b"#bundle\x00".
        self._dgram = dgram
        index = len(_BUNDLE_PREFIX_DGRAM)
        try:
            self._timetag, index = get_timetag(self._dgram, index)
        except OscTypeParseError as e:
            raise OscBundleParseError(
                'Could not get the timetag from datagram') from e
        # Get the contents as a list of OscBundle and OscMessage.
        self._contents = self._parse_contents(index)

    # Return type is actually List[OscBundle], but that would require import annotations from __future__, which is
    # python 3.7+ only.
    def _parse_contents(self, index: int) -> Any:
        contents = []
        try:
            # An OSC Bundle Element consists of its size and its contents.
            # The size is an int32 representing the number of 8-bit bytes in the
            # contents, and will always be a multiple of 4. The contents are either
            # an OSC Message or an OSC Bundle.
            while self._dgram[index:]:
                # Get the sub content size.
                content_size, index = get_int(self._dgram, index)
                # Get the datagram for the sub content.
                content_dgram = self._dgram[index:index + content_size]
                # Increment our position index up to the next possible content.
                index += content_size
                # Parse the content into an OSC message or bundle.
                if OscBundle.dgram_is_bundle(content_dgram):
                    contents.append(OscBundle(content_dgram))
                elif OscMessage.dgram_is_message(content_dgram):
                    contents.append(OscMessage(content_dgram))
                else:
                    _logger.warning('Could not identify content type '
                                    f'of dgram {content_dgram}')
        except (OscTypeParseError, OscMessageParseError, IndexError) as e:
            raise OscBundleParseError(
                "Could not parse a content datagram") from e

        return contents

    @staticmethod
    def dgram_is_bundle(dgram: bytes) -> bool:
        """Returns whether this datagram starts like an OSC bundle."""
        return dgram.startswith(_BUNDLE_PREFIX_DGRAM)

    @property
    def timetag(self) -> int:
        """Returns the timetag associated with this bundle."""
        return self._timetag

    @property
    def num_contents(self) -> int:
        """Shortcut for len(*bundle) returning the number of elements."""
        return len(self._contents)

    @property
    def size(self) -> int:
        """Returns the length of the datagram for this bundle."""
        return len(self._dgram)

    @property
    def dgram(self) -> bytes:
        """Returns the datagram from which this bundle was built."""
        return self._dgram

    def content(self, index) -> Any:
        """Returns the bundle's content 0-indexed."""
        return self._contents[index]

    def __iter__(self) -> Iterator[Any]:
        """Returns an iterator over the bundle's content."""
        return iter(self._contents)


### OSC Bundle Builder ###


class OscBundleBuilder(object):
    """Builds arbitrary OscBundle instances."""

    def __init__(self, timetag: int):
        """
        Build a new bundle with the associated timetag as uint64.
        """
        self._timetag = timetag
        self._contents = []

    def add_content(self, content):
        """Add a new content to this bundle.

        Args:
          - content: Either an OscBundle or an OscMessage
        """
        self._contents.append(content)

    def build(self):
        """Build an OscBundle with the current state of this builder.

        Raises:
          - BuildError: if we could not build the bundle.
        """
        dgram = _BUNDLE_PREFIX_DGRAM
        try:
            dgram += write_timetag(self._timetag)
            for content in self._contents:
                if type(content) is OscMessage or type(content) is OscBundle:
                    size = content.size
                    dgram += write_int(size)
                    dgram += content.dgram
                else:
                    raise OscBundleBuildError(
                        'Content must be either OscBundle or OscMessage '
                        f'found {type(content).__name__}')
            return OscBundle(dgram)
        except OscTypeBuildError as e:
            raise OscBundleBuildError('Could not build the bundle') from e


### OSC Message ###


class OscMessage(object):
    """Representation of a parsed datagram representing an OSC message.

    An OSC message consists of an OSC Address Pattern followed by an OSC
    Type Tag String followed by zero or more OSC Arguments.
    """

    def __init__(self, dgram: bytes) -> None:
        self._dgram = dgram
        self._parameters = []
        self._parse_datagram()

    def _parse_datagram(self) -> None:
        try:
            self._address_regexp, index = get_string(self._dgram, 0)
            if not self._dgram[index:]:
                # No params is legit, just return now.
                return

            # Get the parameters types.
            type_tag, index = get_string(self._dgram, index)
            if type_tag.startswith(','):
                type_tag = type_tag[1:]

            params = []
            param_stack = [params]
            # Parse each parameter given its type.
            for param in type_tag:
                if param == "i":  # Integer.
                    val, index = get_int(self._dgram, index)
                elif param == "f":  # Float.
                    val, index = get_float(self._dgram, index)
                elif param == "d":  # Double.
                    val, index = get_double(self._dgram, index)
                elif param == "s":  # String.
                    val, index = get_string(self._dgram, index)
                elif param == "b":  # Blob.
                    val, index = get_blob(self._dgram, index)
                elif param == "r":  # RGBA.
                    val, index = get_rgba(self._dgram, index)
                elif param == "m":  # MIDI.
                    val, index = get_midi(self._dgram, index)
                elif param == "t":  # osc time tag:
                    val, index = get_timetag(self._dgram, index)
                elif param == "T":  # True.
                    val = True
                elif param == "F":  # False.
                    val = False
                elif param == "[":  # Array start.
                    array = []
                    param_stack[-1].append(array)
                    param_stack.append(array)
                elif param == "]":  # Array stop.
                    if len(param_stack) < 2:
                        raise OscMessageParseError(
                            'Unexpected closing bracket '
                            f'in type tag: {type_tag}')
                    param_stack.pop()
                # TODO: Support more exotic types as described in the specification.
                else:
                    _logger.warning(f'Unhandled parameter type: {param}')
                    continue
                if param not in "[]":
                    param_stack[-1].append(val)
            if len(param_stack) != 1:
                raise OscMessageParseError(
                    f'Missing closing bracket in type tag: {type_tag}')
            self._parameters = params
        except OscTypeParseError as e:
            raise OscMessageParseError(
                'Found incorrect datagram, ignoring it') from e

    @property
    def address(self) -> str:
        """Returns the OSC address regular expression."""
        return self._address_regexp

    @staticmethod
    def dgram_is_message(dgram: bytes) -> bool:
        """Returns whether this datagram starts as an OSC message."""
        return dgram.startswith(b'/')

    @property
    def size(self) -> int:
        """Returns the length of the datagram for this message."""
        return len(self._dgram)

    @property
    def dgram(self) -> bytes:
        """Returns the datagram from which this message was built."""
        return self._dgram

    @property
    def params(self) -> List[Any]:
        """Convenience method for list(self) to get the list of parameters."""
        return list(self)

    def __iter__(self) -> Iterator[float]:
        """Returns an iterator over the parameters of this message."""
        return iter(self._parameters)


### OSC Message Builder ###


class OscMessageBuilder(object):
    """Builds arbitrary OscMessage instances."""

    ARG_TYPE_FLOAT = "f"
    ARG_TYPE_DOUBLE = "d"
    ARG_TYPE_INT = "i"
    ARG_TYPE_STRING = "s"
    ARG_TYPE_BLOB = "b"
    ARG_TYPE_RGBA = "r"
    ARG_TYPE_MIDI = "m"
    ARG_TYPE_TRUE = "T"
    ARG_TYPE_FALSE = "F"
    ARG_TYPE_NIL = "N"

    ARG_TYPE_ARRAY_START = "["
    ARG_TYPE_ARRAY_STOP = "]"

    _SUPPORTED_ARG_TYPES = (
        ARG_TYPE_FLOAT, ARG_TYPE_DOUBLE, ARG_TYPE_INT, ARG_TYPE_BLOB,
        ARG_TYPE_STRING, ARG_TYPE_RGBA, ARG_TYPE_MIDI, ARG_TYPE_TRUE,
        ARG_TYPE_FALSE, ARG_TYPE_NIL)

    def __init__(self, address: str=None) -> None:
        """Initialize a new builder for a message.

        Args:
          - address: The osc address to send this message to.
        """
        self._address = address
        self._args = []

    @property
    def address(self) -> str:
        """Returns the OSC address this message will be sent to."""
        return self._address

    @address.setter
    def address(self, value: str) -> None:
        """Sets the OSC address this message will be sent to."""
        self._address = value

    @property
    def args(self) -> List[Tuple[str, Union[str, bytes, bool, int, float, tuple, list]]]:   # TODO: Make 'tuple' more specific for it is a MIDI packet
        """Returns the (type, value) arguments list of this message."""
        return self._args

    def _valid_type(self, arg_type: str) -> bool:
        if arg_type in self._SUPPORTED_ARG_TYPES:
            return True
        elif isinstance(arg_type, list):
            for sub_type in arg_type:
                if not self._valid_type(sub_type):
                    return False
            return True
        return False

    def add_arg(self, arg_value: Union[str, bytes, bool, int, float, tuple, list], arg_type: str=None) -> None:     # TODO: Make 'tuple' more specific for it is a MIDI packet
        """Add a typed argument to this message.

        Args:
          - arg_value: The corresponding value for the argument.
          - arg_type: A value in ARG_TYPE_* defined in this class,
                      if none then the type will be guessed.
        Raises:
          - ValueError: if the type is not supported.
        """
        if arg_type and not self._valid_type(arg_type):
            raise ValueError(
                f'arg_type must be one of {self._SUPPORTED_ARG_TYPES}, '
                'or an array of valid types')
        if not arg_type:
            arg_type = self._get_arg_type(arg_value)
        if isinstance(arg_type, list):
            self._args.append((self.ARG_TYPE_ARRAY_START, None))
            for v, t in zip(arg_value, arg_type):
                self.add_arg(v, t)
            self._args.append((self.ARG_TYPE_ARRAY_STOP, None))
        else:
            self._args.append((arg_type, arg_value))

    def _get_arg_type(self, arg_value: Union[str, bytes, bool, int, float, tuple, list]) -> str:    # TODO: Make 'tuple' more specific for it is a MIDI packet
        """Guess the type of a value.

        Args:
          - arg_value: The value to guess the type of.
        Raises:
          - ValueError: if the type is not supported.
        """
        if isinstance(arg_value, str):
            arg_type = self.ARG_TYPE_STRING
        elif isinstance(arg_value, (bytes, bytearray, memoryview)):
            arg_type = self.ARG_TYPE_BLOB
        elif arg_value is True:
            arg_type = self.ARG_TYPE_TRUE
        elif arg_value is False:
            arg_type = self.ARG_TYPE_FALSE
        elif isinstance(arg_value, int):
            arg_type = self.ARG_TYPE_INT
        elif isinstance(arg_value, float):
            arg_type = self.ARG_TYPE_FLOAT
        elif isinstance(arg_value, tuple) and len(arg_value) == 4:
            arg_type = self.ARG_TYPE_MIDI
        elif isinstance(arg_value, list):  # NOTE: not a valid case from sc3.
            arg_type = [self._get_arg_type(v) for v in arg_value]
        elif arg_value is None:
            arg_type = self.ARG_TYPE_NIL
        else:
            raise ValueError(
                f'Infered arg_value type is not supported: {type(arg_value)}')
        return arg_type

    def build(self):
        """
        Builds an OscMessage from the current state of this builder.

        Raises:
          - BuildError: if the message could not be build or if the address
            was empty.
        Returns:
          - an OscMessage instance.
        """
        if not self._address:
            raise OscMessageBuildError('OSC addresses cannot be empty')
        dgram = b''
        try:
            # Write the address.
            dgram += write_string(self._address)
            if not self._args:
                dgram += write_string(',')
                return OscMessage(dgram)

            # Write the parameters.
            arg_types = "".join([arg[0] for arg in self._args])
            dgram += write_string(',' + arg_types)
            for arg_type, value in self._args:
                if arg_type == self.ARG_TYPE_STRING:
                    dgram += write_string(value)
                elif arg_type == self.ARG_TYPE_INT:
                    dgram += write_int(value)
                elif arg_type == self.ARG_TYPE_FLOAT:
                    dgram += write_float(value)
                elif arg_type == self.ARG_TYPE_DOUBLE:
                    dgram += write_double(value)
                elif arg_type == self.ARG_TYPE_BLOB:
                    dgram += write_blob(value)
                elif arg_type == self.ARG_TYPE_RGBA:
                    dgram += write_rgba(value)
                elif arg_type == self.ARG_TYPE_MIDI:
                    dgram += write_midi(value)
                elif arg_type in (self.ARG_TYPE_TRUE,
                                  self.ARG_TYPE_FALSE,
                                  self.ARG_TYPE_ARRAY_START,
                                  self.ARG_TYPE_ARRAY_STOP,
                                  self.ARG_TYPE_NIL):
                    continue
                else:
                    raise OscMessageBuildError(
                        f'Incorrect parameter type found {arg_type}')

            return OscMessage(dgram)
        except OscTypeBuildError as e:
            raise OscMessageBuildError(f'Could not build the message') from e


### OSC Packet ###


TimedMessage = collections.namedtuple(
    typename='TimedMessage',
    field_names=('time', 'message'))


class OscPacket():
    def __init__(self, dgram: bytes):
        if OscBundle.dgram_is_bundle(dgram):
            self._messages = self._get_bundle_messages(OscBundle(dgram))
            self._messages = sorted(self._messages, key=lambda x: x.time or 0)
        elif OscMessage.dgram_is_message(dgram):
            self._messages = [TimedMessage(None, OscMessage(dgram))]
        else:
            # Empty packet (because UDP).
            raise OscParseError('OSC packet should at least contain '
                                'an OscMessage or OscBundle')  # *** BUG: ver si un msj con solo address vale

    @property
    def messages(self):
        return self._messages

    def _get_bundle_messages(self, bundle):
        messages = []
        for content in bundle:
            if type(content) is OscMessage:
                messages.append(TimedMessage(bundle.timetag, content))
            else:
                messages.extend(self._get_bundle_messages(content))
        return messages


### OSC Server ###


class UDPHandler(socketserver.BaseRequestHandler):
    def handle(self):  # override
        packet = OscPacket(self.request[0])
        for timed_msg in packet.messages:
            msg = [timed_msg.message.address, *timed_msg.message.params]
            _libsc3.main._osc_interface.recv(
                self.client_address, timed_msg.time, *msg)
        # NOTE: Exception are handled by BaseServer.handle_error, "The default action is to print the traceback to standard error and continue handling further requests."
        # NOTE: "The type of self.request is different for datagram or stream services. For stream services, self.request is a socket object; for datagram services, self.request is a pair of string and socket."


class OSCUDPServer(socketserver.UDPServer):
    def verify_request(self, request, client_address):  # override
        return (OscBundle.dgram_is_bundle(request[0]) or
                OscMessage.dgram_is_message(request[0]))

    def handle_error(self, request, client_address):  # override
        _logger.error(
            'Exception happened during processing '
            f'request from {client_address}',
            exc_info=sys.exc_info())


# class ThreadingOSCUDPServer(socketserver.ThreadingMixIn, OSCUDPServer):
#     pass



"""Env.sc"""

import copy
import operator

from ..base import main as _libsc3
from ..base import utils as utl
from ..base import builtins as bi
from . import _graphparam as gpp
from .ugens import trig as trg
from .ugens import oscillators as ocl


__all__ = ['Env']


class Env(gpp.UGenParameter, gpp.NodeParameter):
    _SHAPE_NAMES = {
        'step': 0,
        'lin': 1,
        'linear': 1,
        'exp': 2,
        'exponential': 2,
        'sin': 3,
        'sine': 3,
        'wel': 4,
        'welch': 4,
        'sqrt': 6,
        'squared': 6,
        'cub': 7,
        'cubed': 7,
        'hold': 8
    }

    def __init__(self, levels=None, times=None, curves='lin',
                 release_node=None, loop_node=None, offset=0):
        super(gpp.UGenParameter, self).__init__(self)
        self.levels = levels or [0, 1, 0]  # Can't be empty or zero either.
        self.times = utl.wrap_extend(
            utl.as_list(times or [1, 1]), len(self.levels) - 1)
        self.curves = curves
        self.release_node = release_node
        self.loop_node = loop_node
        self.offset = offset
        self._envgen_format = None
        self._interpolation_format = None

    # no newClear
    # no kr
    # no ar
    # no setters

    # // Methods to make some typical shapes.

    # // Fixed duration envelopes.

    @classmethod
    def xyc(cls, xyc):
        if any(len(i) != 3 for i in xyc):
            raise ValueError('xyc list must contain only sequences of length 3')
        xyc = xyc[:]  # Ensures internal state.
        xyc.sort(key=lambda x: x[0])
        times, levels, curves = utl.flop(xyc)
        offset = times[0]
        times = [b - a for a, b in utl.pairwise(times)]  # differentiate
        curves.pop(-1)
        return cls(levels, times, curves, offset=offset)

    @classmethod
    def pairs(cls, pairs, curves=None):
        pairs = pairs[:]  # Ensures internal state.
        if curves is None:
            for i in range(len(pairs)):
                pairs[i].append('lin')
        elif isinstance(curves, str):
            for i in range(len(pairs)):
                pairs[i].append(curves)
        else:
            if len(pairs) != len(curves):
                raise ValueError('pairs and curves must have the same length')
            for i in range(len(pairs)):
                pairs[i].append(curves[i])
        return cls.xyc(pairs)

    @classmethod
    def triangle(cls, dur=1.0, level=1.0):
        dur *= 0.5
        return cls([0, level, 0], [dur, dur])

    @classmethod
    def sine(cls, dur=1.0, level=1.0):
        dur *= 0.5
        return cls([0, level, 0], [dur, dur], 'sine')

    @classmethod
    def perc(cls, attack_time=0.01, release_time=1.0, level=1.0, curve=-4.0):
        return cls([0, level, 0], [attack_time, release_time], curve)

    @classmethod
    def linen(cls, attack_time=0.01, sustain_time=1.0, release_time=1.0,
              level=1.0, curve='lin'):
        return cls(
            [0, level, level, 0],
            [attack_time, sustain_time, release_time], curve)

    @classmethod
    def step(cls, levels=None, times=None, release_node=None,
             loop_node=None, offset=0):
        levels = levels or [0, 1]
        times = times or [1, 1]
        if len(levels) != len(times):
            raise ValueError('levels and times must have same length')
        levels = levels[:]  # Ensures internal state.
        levels.insert(0, levels[0])
        return Env(levels, times, 'step', release_node, loop_node, offset)

    # // Envelopes with sustain.

    @classmethod
    def cutoff(cls, release_time=0.1, level=1.0, curve='lin'):
        curve_no = cls._shape_number(curve)
        release_level = bi.dbamp(-100) if curve_no == 2 else 0
        return cls([level, release_level], [release_time], curve, 0)

    @classmethod
    def dadsr(cls, delay_time=0.1, attack_time=0.01, decay_time=0.3,
              sustain_level=0.5, release_time=1.0, peak_level=1.0,
              curve=-4.0, bias=0.0):
        return cls(
            utl.list_binop(
                operator.add,
                [0, 0, peak_level, peak_level * sustain_level, 0], bias),
            [delay_time, attack_time, decay_time, release_time], curve, 3)

    @classmethod
    def adsr(cls, attack_time=0.01, decay_time=0.3, sustain_level=0.5,
             release_time=1.0, peak_level=1.0, curve=-4.0, bias=0.0):
        return cls(
            utl.list_binop(
                operator.add,
                [0, peak_level, peak_level * sustain_level, 0], bias),
            [attack_time, decay_time, release_time], curve, 2)

    @classmethod
    def asr(cls, attack_time=0.01, sustain_level=1.0,
            release_time=1.0, curve=-4.0):
        return cls(
            [0, sustain_level, 0], [attack_time, release_time], curve, 1)

    @classmethod
    def cyclic(cls, levels, times, curves='lin'):  # was *circle
        times = utl.wrap_extend(utl.as_list(times), len(levels))
        last_time = times.pop()
        curves = utl.wrap_extend(utl.as_list(curves), len(levels))
        last_curve = curves.pop()
        return cls(levels, times, curves).circle(last_time, last_curve)

    def circle(self, last_time=0.0, last_curve='lin'):
        # // Connect releaseNode (or end) to first node of envelope.
        if _libsc3.main._current_synthdef is None:
            raise Exception('circle can only be used within graph functions')
        first_0_then_1 = trg.Latch.kr(1.0, ocl.Impulse.kr(0.0))
        if self.release_node is None:
            self.levels = [0.0, *self.levels, 0.0]
            self.curves = ult.wrap_extend(
                utl.as_list(self.curves), len(self.times))
            self.curves = [last_curve, *self.curves, 'lin']
            self.times = [
                first_0_then_1 * last_time, *self.times, float('inf')]
            self.release_node = len(self.levels) - 2
        else:
            self.levels = [0.0, *self.levels]
            self.curves = ult.wrap_extend(
                utl.as_list(self.curves), len(self.times))
            self.curves = [last_curve, *self.curves]
            self.times = [first_0_then_1 * last_time, *self.times]
            self.release_node += 1
        self.loop_node = 0
        return self

    @property
    def duration(self):
        return utl.list_sum(self.times)

    @duration.setter
    def duration(self, value):
        res = utl.list_binop(
            operator.mul, self.times, 1 / self.total_duration())
        self.times = utl.list_binop(operator.mul, res, value)

    def total_duration(self):
        duration = utl.list_sum(self.times)
        return utl.list_max(utl.as_list(duration))

    @property
    def release_time(self):
        if self.release_node is None:
            return 0.0
        else:
            return utl.list_sum(self.times[self.release_node:])

    @property
    def is_sustained(self):
        return self.release_node is not None

    def range(self, lo=0.0, hi=1.0):
        obj = copy.copy(self)
        min = utl.list_min(obj.levels)
        max = utl.list_max(obj.levels)
        obj.levels = utl.list_narop(bi.linlin, obj.levels, min, max, lo, hi)
        return obj

    def exprange(self, lo=0.01, hi=1.0):
        obj = copy.copy(self)
        min = utl.list_min(obj.levels)
        max = utl.list_max(obj.levels)
        obj.levels = utl.list_narop(bi.linexp, obj.levels, min, max, lo, hi)
        return obj

    def curverange(self, lo=0.0, hi=1.0, curve=-4):
        obj = copy.copy(self)
        min = utl.list_min(obj.levels)
        max = utl.list_max(obj.levels)
        obj.levels = utl.list_narop(
            bi.lincurve, obj.levels, min, max, lo, hi, curve)
        return obj

    # TODO
    # asMultichannelSignal
    # asSignal
    # discretize
    # storeArgs
    # ==
    # hash
    # at
    # embedInStream
    # asStream
    # asPseg
    # blend
    # delay
    # circle (moved up)
    # test

    @classmethod
    def _shape_number(cls, name):
        name = utl.as_list(name)
        ret = []
        for item in name:
            if gpp.ugen_param(item)._is_valid_ugen_input():
                ret.append(5)  # 'curvature value', items is not NaN SimpleNumber.
            else:
                try:
                    shape = cls._SHAPE_NAMES[item]
                    ret.append(shape)
                except KeyError as e:
                     raise ValueError(f"invalid Env shape '{item}'") from e
        return utl.unbubble(ret)

    @classmethod
    def _curve_value(cls, curve):
        if isinstance(curve, list):
            ret = []
            for item in curve:
                if gpp.ugen_param(item)._is_valid_ugen_input():
                    ret.append(item)
                else:
                    ret.append(0)
            return ret
        else:
            if gpp.ugen_param(curve)._is_valid_ugen_input():
                return curve
            else:
                return 0

    def envgen_format(self):  # Was asMultichannelArray.
        if self._envgen_format:  # this.array
            return self._envgen_format

        # prAsArray
        levels = gpp.ugen_param(self.levels)._as_ugen_input()
        times = gpp.ugen_param(self.times)._as_ugen_input()
        curves = gpp.ugen_param(utl.as_list(self.curves))._as_ugen_input()
        size = len(self.times)
        contents = []

        contents.append(levels[0])
        contents.append(size)
        aux_input = gpp.ugen_param(self.release_node)._as_ugen_input()
        if aux_input is None:
            aux_input = -99
        contents.append(aux_input)
        aux_input = gpp.ugen_param(self.loop_node)._as_ugen_input()
        if aux_input is None:
            aux_input = -99
        contents.append(aux_input)

        for i in range(size):
            contents.append(levels[i + 1])
            contents.append(times[i])
            contents.append(type(self)._shape_number(curves[i % len(curves)]))
            contents.append(type(self)._curve_value(curves[i % len(curves)]))

        self._envgen_format = [tuple(i) for i in utl.flop(contents)]
        return self._envgen_format

    def interpolation_format(self):
        '''This version is for IEnvGen which has a special format.'''
        if self._interpolation_format:
            return self._interpolation_format

        levels = gpp.ugen_param(self.levels)._as_ugen_input()
        times = gpp.ugen_param(self.times)._as_ugen_input()
        curves = gpp.ugen_param(utl.as_list(self.curves))._as_ugen_input()
        size = len(self.times)
        contents = []

        aux_input = gpp.ugen_param(self.offset)._as_ugen_input()
        if aux_input is None:
            aux_input = 0
        contents.append(aux_input)
        contents.append(levels[0])
        contents.append(size)
        contents.append(utl.list_sum(times))

        for i in range(size):
            contents.append(times[i])
            contents.append(type(self)._shape_number(curves[i % len(curves)]))
            contents.append(type(self)._curve_value(curves[i % len(curves)]))
            contents.append(levels[i + 1])

        self._interpolation_format = [tuple(i) for i in utl.flop(contents)]
        return self._interpolation_format


    ### Node parameter interface ###

    def _as_control_input(self):
        return utl.unbubble(self.envgen_format())

    def _embed_as_osc_arg(self, lst):
        gpp.node_param(self._as_control_input())._embed_as_osc_arg(lst)



"""AbstractFunction.sc"""

import inspect

from . import absobject as aob
from . import builtins as bi


__all__ = ['value', 'Function', 'function']


def value(obj, *args, **kwargs):
    '''
    Utility function for optional value/function parameters like
    completion_msg. If obj is a function it gets evaluated with `*args`
    and `**kwargs` and the result is returned, else obj is returned as is.
    Spare parameters are discarded.
    '''
    if callable(obj):
        parameters = inspect.signature(obj).parameters
        if any(p.kind == p.VAR_POSITIONAL for p in parameters.values()):
            return obj(*args, **kwargs)
        else:
            nargs = len(parameters)
            return obj(*args[:nargs], **kwargs)
    else:
        return obj


class AbstractFunction(aob.AbstractObject):
    def __call__(self, *args):
        raise NotImplementedError(
            f'{type(self).__name__} does not implement __call__')


    ### AbstractObject interface ###

    def _compose_unop(self, selector):
        return UnaryOpFunction(selector, self)

    def _compose_binop(self, selector, other):
        return BinaryOpFunction(selector, self, other)

    def _rcompose_binop(self, selector, other):
        return BinaryOpFunction(selector, other, self)

    def _compose_narop(self, selector, *args):
        return NAryOpFunction(selector, self, *args)

    # applyTo
    # <> function composition

    def sampled(self, n=80, from_=0.0, to=1.0):
        range_ = to - from_
        step = range_ / (n - 1)
        end = range_ / step + 1
        start = int(from_)
        offset = from_ - start
        values = []
        for i in range(start, int(end)):
            values.append(self(i * step + offset))

        def sampled_func(x):
            pos = (bi.clip(x, from_, to) - from_) / (to - from_) * (n - 1)
            i = bi.ceil(pos) - 1  # i = int(bi.roundup(pos)) - 1  # ceil(x) -> int
            return bi.blend(values[i], values[i + 1], bi.absdif(pos, i))  # uses blendAt from Object.

        return sampled_func

    # ### UGen graph parameter interface ###
    # # Is a mistake to make AbstractFunction an UGenParameter because some
    # # subclases are not (e.g. every stream), if needed it could be better to
    # # create a ugen_param for functions, callables or alike.
    #
    # def _is_valid_ugen_input(self):
    #     return True
    #
    # def _as_ugen_input(self, *ugen_cls):
    #     return self(*ugen_cls)
    #
    # def _as_control_input(self):  # Is NodeParameter interface.
    #     return self()
    #
    # def _as_audio_rate_input(self, *args):
    #     res = self(*args)
    #     if gpp.ugen_param(res)._as_ugen_rate() != 'audio':
    #         return xxx.K2A.ar(res)
    #     return res


class UnaryOpFunction(AbstractFunction):
    def __init__(self, selector, a):
        self.selector = selector
        self.a = a

    def __call__(self, *args, **kwargs):
        return self.selector(self.a(*args, **kwargs))


class BinaryOpFunction(AbstractFunction):
    def __init__(self, selector, a, b):
        self.selector = selector
        self.a = a
        self.b = b

    def __call__(self, *args, **kwargs):
        a_value = self.a(*args, **kwargs) if callable(self.a) else self.a
        b_value = self.b(*args, **kwargs) if callable(self.b) else self.b
        return self.selector(a_value, b_value)


class NAryOpFunction(AbstractFunction):
    def __init__(self, selector, a, *args):
        self.selector = selector
        self.a = a
        self.args = args

    def __call__(self, *args, **kwargs):
        evaluated_args = [
            x(*args, **kwargs) if isinstance(x, Function)
            else x for x in self.args]
        return self.selector(self.a(*args, **kwargs), *evaluated_args)


# class FunctionList(AbstractFunction):
#     ...


### Function.sc ###

class Function(AbstractFunction):
    def __init__(self, func):
        if inspect.isfunction(func):
            parameters = inspect.signature(func).parameters
            self._nargs = len(parameters)
            self._kwords = parameters.keys()
            self.func = func
            self._clock = None  # Internal compatibility as Clock item.
        else:
            raise TypeError(
                'Function wrapper only apply to user-defined functions')

    def __call__(self, *args, **kwargs):
        kwargs = {k: kwargs[k] for k in kwargs.keys() & self._kwords}
        return self.func(*args[:self._nargs], **kwargs)

    def __awake__(self, beats, seconds, clock):  # Function, Routine, PauseStream, (Nil, Object).
        return self.func(*(beats, seconds, clock)[:self._nargs])


# decorator syntax
def function(func):
    return Function(func)


# Thunk
# UGenThunk
